/*
	$Id: ArrayBlockTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005 Jari Hakkinen, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArrayDesignBlock;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BlockInfo;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FeatureBatcher;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.Well;
import net.sf.basedb.core.data.FeatureData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.Map;

/**
<dl>
<dt><b>Base 1 tables</b></dt>
<dd>ArrayBlock</dd>
<dt><b>Unused columns</b></dt>
<dd>featureDiameter</dd>
<dt><b>Notes</b></dt>
<dd>
<ul>
<li>ArrayType is named ArrayDesign in BASE 2.</li>
</ul>
</dd>
</dl>

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class ArrayBlockTransfer extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.ArrayBlockTransfer");
	
	private ArrayTypeTransfer arrayTypeT = null;
	private MappedPlateTransfer mappedPlateT = null;
	private ReporterTransfer reporterT = null;

	/**
	@param arrayTypeT A finished {@link ArrayTypeTransfer}
	 */
	ArrayBlockTransfer(ArrayTypeTransfer arrayTypeT,
			MappedPlateTransfer mappedPlateT, ReporterTransfer reporterT)
	{
		log.debug("Checking that necessary transfers have been done before proceeding with ArrayBlocks");
		if (arrayTypeT != null && arrayTypeT.isDone() && mappedPlateT != null
				&& mappedPlateT.isDone() && reporterT != null
				&& reporterT.isDone())
		{
			this.arrayTypeT = arrayTypeT;
			this.mappedPlateT = mappedPlateT;
			this.reporterT = reporterT;
		}
		else
		{
			if (!arrayTypeT.isDone())log.debug("Dependency: ArrayTypes haven't been transfered");
			if (!mappedPlateT.isDone()) log.debug("Dependency: MappedPlates haven't been transfered");
			if (!reporterT.isDone()) log.debug("Dependency: Reporters haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring ArrayBlocks...");
		startProgress(getInt("countFeatures"), "ArrayBlocks");
		DbControl dc = null;
		for (Map.Entry<Integer, Integer> entry : arrayTypeT.getIdMap().entrySet())
		{
			cleanUpMemory();
			dc = newDbControl();
			Integer base2Id = entry.getValue();
			Integer base1Id = entry.getKey();
			
			if (base2Id != null)
			{
				try
				{
					ArrayDesign arrayDesign = ArrayDesign.getById(dc, base2Id);
					PreparedStatement ps = prepareStatementFromFile("selectArrayBlocksByArrayType");
					FeatureBatcher featureBatcher = 
						arrayDesign.getFeatureBatcher(FeatureIdentificationMethod.COORDINATES, null);
					log.debug("Setting batch size to: " + getBatchSize());
					featureBatcher.setBatchSize(getBatchSize());
					addArrayDesignBlock(
							dc, arrayDesign, base1Id, ps, 0, getBatchSize(),
							featureBatcher);
					featureBatcher.flushInsert();
					featureBatcher.close();
				}
				catch (ItemNotFoundException e)
				{
					log.warn("ArrayDesign with id " + base2Id
							+ " was not found");
					if (log.isDebugEnabled()) log.debug(e);
				}
				catch (Exception e)
				{
					log.error("FAILED transferring ArrayBlocks", e);
					stop();
				}
			}
			else
			{
				log.warn("Id for ArrayDesign is: " + base2Id);
			}
			commit(dc);
		}
		done();
		log.info("Transferring ArrayBlocks...DONE");
	}

	/**
	Adds an ArrayDesignBlock to a given ArrayDesign
	
	@param dc {@link DbControl} used to create a new ArrayDesignBlock
	@param arrayDesign <code>ArrayDesign</code>
	@param base1ArrayDesignId Base 1 ArrayDesign Id
	 */
	private void addArrayDesignBlock(DbControl dc, ArrayDesign arrayDesign,
			int base1ArrayDesignId, PreparedStatement ps, int from, int size,
			FeatureBatcher featureBatcher)
	{
		log.info("Adding ArrayDesignBlocks");
		try
		{
			ps.setInt(1, base1ArrayDesignId);
			ps.setInt(2, from);
			ps.setInt(3, size);
			ResultSet arrayBlock = ps.executeQuery();
			int rowCounter = 0;
			while (arrayBlock.next())
			{
				rowCounter++;
				int blockNum = arrayBlock.getInt(2);
				int metaRow = arrayBlock.getInt(10);
				int metaColumn = arrayBlock.getInt(11);
				BlockInfo blockInfo = new BlockInfo(blockNum, metaRow,
						metaColumn);
				if (blockInfo != null)
				{
					log.debug("Setting block info for block number: " + blockNum);
					ArrayDesignBlock arrayDesignBlock = arrayDesign.addArrayDesignBlock(blockInfo);
					arrayDesignBlock.setOriginX(arrayBlock.getInt(3));
					arrayDesignBlock.setOriginY(arrayBlock.getInt(4));
					// featureDiameter is not used in BASE 2
					arrayDesignBlock.setBlockSizeX(arrayBlock.getInt(6));
					arrayDesignBlock.setSpacingX(arrayBlock.getInt(7));
					arrayDesignBlock.setBlockSizeY(arrayBlock.getInt(8));
					arrayDesignBlock.setSpacingY(arrayBlock.getInt(9));
					PreparedStatement ps1 = prepareStatementFromFile("selectFeaturesByArrayTypeAndBlock");
					ps1.setInt(1, base1ArrayDesignId);
					ps1.setInt(2, blockNum);
					log.debug("Adding features to block number: " + blockNum);
					addFeatures(
							dc, arrayDesignBlock, ps1, 0, getBatchSize(), featureBatcher);
				}
				else
				{
					log.debug("BlockInfo is null for block number: " + blockNum);
				}
			}
			arrayBlock.close();
			if (rowCounter == size)
			{
				addArrayDesignBlock(dc, arrayDesign, base1ArrayDesignId, ps, from
						+ size, size, featureBatcher);
			}			
		}		
		catch (Exception e)
		{
			log.error("FAILED to add ArrayDesignBlock", e);
			stop();
		}
		log.info("ArrayDesignBlocks are now added");
	}

	/**
	This recursive method adds features to an ArrayDesignBlock.
	
	@param dbc {@link DbControl} used to transfer this item
	@param arrayDesignBlock <code>ArrayDesignBlock</code> from which to
	       create FeatureData
	@param ps PreparedStatement used to fetch features from Base 1
	@param from Position in the result set to start reading from. Normally
	       this is 0 when you call it.
	@param size Number of rows to read from specified position
	@param featureBatcher FeatureBatcher used for saving this set of
	       FeatureData
	 */
	private void addFeatures(DbControl dbc, ArrayDesignBlock arrayDesignBlock,
			PreparedStatement ps, int from, int size,
			FeatureBatcher featureBatcher)
	{
		log.info("Adding features to array design block: " + arrayDesignBlock.getBlockNumber());
		try
		{
			ps.setInt(3, from);
			ps.setInt(4, size);
			ResultSet rs = ps.executeQuery();
			int rowCounter = 0;
			while (rs.next())
			{
				rowCounter++;
				FeatureData featureData = null;
				int reporterId = rs.getInt(7);
				int wellId = rs.getInt(8);
				int plateId = rs.getInt(9);
				if (wellId != 0)
				{
					try
					{
						Plate plate = mappedPlateT.getPlateById(dbc, plateId);
						int col = rs.getInt(10) - 1;
						int row = rs.getInt(11) - 1;
						Well well = plate.getWell(row, col);
						if (well != null)
						{
							featureData = featureBatcher.newFeature(
									arrayDesignBlock, well);
						}
					}
					catch (BaseException e)
					{
						log.warn("Couldn't get feature data");
						if (log.isDebugEnabled()) log.debug("\t" + e);
					}
				}
				else
				{
					featureData = featureBatcher.newFeature(
						arrayDesignBlock, reporterT.getReporterById(dbc, reporterId));
				}
				if (featureData != null)
				{
					featureData.setColumn(rs.getInt(4));
					featureData.setRow(rs.getInt(5));
					featureBatcher.insert(featureData);
					if (log.isDebugEnabled())
					{
						log.debug("Inserting feature data: column " + rs.getInt(4) + ", row: " + rs.getInt(5));
					}
				}
				progress.increase();
			}
			rs.close();
			if (rowCounter == size)
			{
				addFeatures(
						dbc, arrayDesignBlock, ps, from + size, size,
						featureBatcher);
			}
		}
		catch (Exception e)
		{
			log.error("FAILED adding features", e);
			stop();
		}
	}
}
