/*
	$Id: RawBioAssayDataTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005 Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataBatcher;
import net.sf.basedb.core.data.RawData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
	This class provides methods for transferring RawBioAssay and RawBioAssayData
	from base 1 to BASE 2. Base 1 files are located in /userfiles/rawdata/
	
	@author Gregory, Martin
	@version 2.0
	@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class RawBioAssayDataTransfer 
	extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.RawBioAssayDataTransfer");
	
	/**
	Set string and text properties, order must match the select statement
	except for the id and reporterId in the beginning Extended properties are
	defined in the extended-properties.xml file for class ReporterData Why do
	it this way and not using the column name for lookup? Performance is
	increased by using the column index instead of name.
	 */
	private String[] floatProperties = { "diameter", "ch1FgMedian",
			"ch1FgMean", "ch1FgSd", "ch1BgMedian", "ch1BgMean", "ch1BgSd",
			"ch2FgMedian", "ch2FgMean", "ch2FgSd", "ch2BgMedian", "ch2BgMean",
			"ch2BgSd", "ratiosSd", "rgnRatio", "rgnR2", "mValue", "CV" };
	private String[] intProperties = { "ch1PercSd1", "ch1PercSd2",
			"ch1PercSat", "ch2PercSd1", "ch2PercSd2", "ch2PercSat", "fgPixels",
			"bgPixels", "flags", };
	private RawBioAssayTransfer rbaT = null;
	private ReporterTransfer reporterT = null;

	/**
	Constructs a new RawBioAssayTransfer.
	
	@param rbaT A finished {@link RawBioAssayTransfer}
	@param reporterT A finished {@link ReporterTransfer}
	 */
	RawBioAssayDataTransfer(RawBioAssayTransfer rbaT, ReporterTransfer reporterT)
	{
		super();
		log.debug("Checking that necessary transfers have been done before proceeding with RawBioAssayData");
		if (rbaT != null && rbaT.isDone() && 
			reporterT != null && reporterT.isDone())
		{
			this.rbaT = rbaT;
			this.reporterT = reporterT;
		}
		else
		{
			if (!rbaT.isDone()) log.debug("Dependency: RawBioAssays haven't been transfered");
			if (!reporterT.isDone()) log.debug("Dependency: Reporters haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring RawBioAssayData...");
		int count = getInt("countRawBioAssayData");
		startProgress(count, "RawBioAssaysData");
		runUnBatched(
				prepareStatementFromFile("selectRawBioAssaysWithSpotData"), 0,
				getBatchSize());
		progress.stop();
		log.info("Transferring RawBioAssayData...DONE");
	}

	/**
	Transfers associated RawBioAssayData for each RawBioAssay and creates
	spotImages if necessary.
	
	@param rs <code>ResultSet</code> containing the data for this item.
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected void createItem(ResultSet rs) throws SQLException
	{
		int rawBioAssayId = rs.getInt(1);
		DbControl dc = newDbControl();
		RawBioAssay rawBioAssay = rbaT.getRawBioAssayById(dc, rawBioAssayId);		
		if (rawBioAssay != null)
		{
			if (log.isDebugEnabled()) log.debug("Starts to transfer RawBioAssayData for RawBioAssay:" + rawBioAssay.getName());
			ArrayDesign ad = rawBioAssay.getArrayDesign();
			boolean hasArrayDesign = ad != null ? true : false;
			if (log.isDebugEnabled())
			{
				log.debug("RawBioAssay '" + rawBioAssayId + "' " + (hasArrayDesign ? "has " : "no ") + "Array Design");
			}
			if (!rawBioAssay.hasData())
			{
				transferRawBioAssayData(rawBioAssay, rawBioAssayId, hasArrayDesign);
			}
			else
			{
				log.info("Skipping since it already has data");
				progress.increase(rawBioAssay.getNumDbSpots());
			}
			progress.decrease(); // We do not want to count the RawBioAssays
			if (log.isDebugEnabled())
			{
				log.debug("Transfered RawBioAssayData for RawBioAssay:" + rawBioAssay.getName());
				log.debug("--------------------------------------------------------------------------");
			}
		}
		else
		{
			log.warn("No RawBioAssay with id '" + rawBioAssayId + "' could be found");
		}
		commit(dc);
	}

	/**
	This method starts the data transfer. Defining the SQL statement used for
	data selection.
	
	@param rawBioAssay Existing rawBioAssay for which data will be
	       transfered.
	@param rawBioAssayId Base 1 id of the rawBioAssay.
	 */
	private void transferRawBioAssayData(RawBioAssay rawBioAssay,
			int rawBioAssayId, boolean hasArrayDesign)
	{
		/*
	The order of the columns is important, they have to match the order
	floatProperties[] followed by intProperties[]
		 */
		cleanUpMemory();
		RawDataBatcher batcher = rawBioAssay.getRawDataBatcher(null, null);
		batcher.setBatchSize(getBatchSize());
		runRawBioAssay(rawBioAssay, batcher, hasArrayDesign, rawBioAssayId);
		batcher.flush();
		batcher.close();
	}

	/**
	Recursive run over a SQL query defined by a prepared statement and LIMIT
	values from and size.
	
	@param rawBioAssay The raw bio assay to transfer data to
	@param batcher RawDataBatcher used in this transfer
	@param hasArrayDesign (Currently not used)
	@param base1RawBioAssayId The ID of the raw bioassay in the BASE 1.2 installation
	@see Transfer#prepareStatementFromFile(String)
	 */
	protected void runRawBioAssay(RawBioAssay rawBioAssay,  
			RawDataBatcher batcher, boolean hasArrayDesign, int base1RawBioAssayId)
	{
		try
		{
			int rowCount = 0;
			ResultSet rs = null;
			Throwable t = null;
			int numTries = 0;
			log.debug(QUICK_TRANSFER ? "Minimum RawBioAssayData is transfered" : "Full transfer of RawBioAssayData");
			// Fix for unstable BASE 1 server that is randomly losing the connection
			// For this to work the connection must be created with autoReconnect=true
			while (rs == null && numTries < 2)
			{
				numTries++;
				try
				{					
					PreparedStatement ps = QUICK_TRANSFER ? prepareStatementFromFile("selectMinimumRawBioAssayData")
						: prepareStatementFromFile("selectRawBioAssayData");				
					
					ps.setInt(1, base1RawBioAssayId);
					rs = ps.executeQuery();
					t = null;
				}
				catch (Throwable t2)
				{
					t = t2;
					log.warn("Exception when loading BASE 1 data for raw bioassay: " + rawBioAssay);					
					if (log.isDebugEnabled()) 
					{
						log.debug(t2);
						log.debug("...sleeping 2 seconds");
					}
					try
					{
						Thread.sleep(2000);
					}
					catch (InterruptedException ex)
					{
						log.debug("Something waked me up", ex);
					}
				}
			}
			if (rs == null || t != null) throw new BaseException(t);
			DbControl dc = batcher.getDbControl();
			while (rs.next())
			{
				RawData rd = batcher.newRawData();
				rd.setReporter(reporterT.getReporterById(dc, rs.getInt(2)));
				rd.setBlock(rs.getInt(3));
				rd.setColumn(rs.getInt(4));
				rd.setRow(rs.getInt(5));
				rd.setX(rs.getFloat(6));
				rd.setY(rs.getFloat(7));
				if (!QUICK_TRANSFER)
				{
					int colIndex = 8; // Extended properties start at the 8:th
										// column
					for (String property : floatProperties)
					{
						rd.setExtended(property, rs.getFloat(colIndex));
						colIndex++;
					}
					for (String property : intProperties)
					{
						rd.setExtended(property, rs.getInt(colIndex));
						colIndex++;
					}
				}
				try
				{
					batcher.insert(rd);					
				}
				catch (Exception e)
				{
					log.debug("Couldn't insert the RawData", e); // No such feature
				}
				progress.increase();
				rowCount++;
			}
			if (log.isInfoEnabled())
			{
				log.info("Transfered " + rowCount + " rows of RawData for raw bioassay: " + rawBioAssay.getName());
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED executing SQL statement", e);
			stop();
		}
		catch (BaseException e)
		{
			log.warn("Couldn't get RawData from BASE1");
			log.debug("\t" + e);
			progress.increase();
		}
	}

}
