/*
	$Id: Button.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import net.sf.basedb.clients.web.taglib.Page;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

/**
	This tag creates a button on a toolbar. The tag must be located inside
	a {@link Toolbar &lt;tbl:toolbar&gt;}
	tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:button
      id=...
      clazz=...
      style=...
      image=...
      imagealign=left|right
      onclick=...
      title=...
      tooltip=...
      disabled=true|false
      visible=true|false
   &gt;
</pre>

	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>
			The ID of the button. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
button = document.getElementById('&lt;id&gt;');
</pre>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>button</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>image</td>
		<td>-</td>
		<td>no</td>
		<td>
			The file name of an image to show on the button. It is expected
			that the image is located in the <code>/images</code> directory.
			Do not specify a complete path.
		</td>
	</tr>
	<tr>
		<td>imagealign</td>
		<td>left</td>
		<td>no</td>
		<td>
			If the image should be aligned to the left or right of the text.
		</td>
	</tr>
	<tr>
		<td>onclick</td>
		<td>-</td>
		<td>no</td>
		<td>
			A JavaScript statement that is executed when the user click on
			the button.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			The text of the button.
		</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td>-</td>
		<td>no</td>
		<td>
			A text that will be shown while the user has the mouse over the button.
		</td>
	</tr>
	<tr>
		<td>disabled</td>
		<td>-</td>
		<td>false</td>
		<td>
			If the button is disabled or not. The values '0', 'no' or 'false' will enabled the
			button, all other values will disable it.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the button should be visible or not.
		</td>
	</tr>
	</table>
	<p>
	This tag has no subtags.

	@author Nicklas
	@version 2.0
*/
public class Button
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = -6294795199053088632L;

	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	/**
		The parent &lt;tbl:toolbar&gt; tag.
	*/
	private Toolbar toolbar;

	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "button";
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;
	
	/**
		An optional image to show in the button.
	*/
	private String image = null;
	
	/**
		LEFT or RIGTH;
	*/
	private String imageAlign = "left";
	
	/**
		A JavaScript statement that is executet when the button is clicked.
	*/
	private String onClick = null;
	
	/**
		An optional title text on the button.
	*/
	private String title = null;
	
	/**
		An optional tooltip that will show when the mouse is over the button.
	*/
	private String tooltip = null;
	
	/**
		IF the button is disabled or not.
	*/
	private boolean disabled = false;

	/**
		If the button should be visible or not.
	*/
	private boolean visible = true;

	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}

	public void setImage(String image)
	{
		this.image = image;
	}
	public String getImage()
	{
		return image;
	}
	
	public void setImagealign(String imageAlign)
	{
		this.imageAlign = imageAlign;
	}
	public String getImagealign()
	{
		return imageAlign;
	}

	public void setOnclick(String onClick)
	{
		this.onClick = onClick;
	}
	public String getOnclick()
	{
		return onClick;
	}

	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}
	
	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	public void setDisabled(boolean disabled)
	{
		this.disabled = disabled;
	}
	public boolean isDisabled()
	{
		return disabled;
	}
	
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	/*
		Display methods
	*/
	public int doStartTag()
		throws JspException
	{
		if (!isVisible()) return SKIP_BODY;
		if (!(getParent() instanceof Toolbar)) throw new JspTagException("Tag <tbl:button> must be inside a <tbl:toolbar> tag");
		toolbar = (Toolbar)getParent();
		table = toolbar.getTable();
		Page page = table != null ? table.getPage() : (Page)findAncestorWithClass(this, Page.class);

		boolean displayImage = getImage() != null && toolbar.hasImages();
		boolean displayText = toolbar.hasText() || !displayImage;
		
		StringBuilder sb = new StringBuilder();
		sb.append("<td");
		if (getId() != null) sb.append(" id=\"").append(getId()).append("\"");
		if (getClazz() != null)
		{
			sb.append(" class=\"").append(getClazz()).append(isDisabled() ? "_disabled" : "").append("\"");
			if (!isDisabled())
			{
				sb.append(" onmouseover=\"this.className='").append(getClazz()).append("_hover';\"");
				sb.append(" onmouseout=\"this.className='").append(getClazz()).append("';\"");
			}
		}
		if (getOnclick() != null && !isDisabled()) sb.append(" onclick=\"").append(getOnclick()).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
		String tooltip = getTooltip();
		if (!displayText && tooltip == null) tooltip = getTitle();
		if (tooltip != null) sb.append(" title=\"").append(tooltip).append("\"");
		sb.append(">");
		if (displayImage)
		{
			sb.append("<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\"><tr>");
			if ("left".equalsIgnoreCase(getImagealign()))
			{
				sb.append("<td><img src=\"").append(page.getRoot()).append("images/").append(getImage()).append("\"");
				sb.append(" border=\"0\"></td>");
			}
			sb.append("<td>");
		}
		if (displayText && getTitle() != null) sb.append("&nbsp;").append(getTitle()).append("&nbsp;");
		if (displayImage)
		{
			sb.append("</td>");
			if (!"left".equalsIgnoreCase(getImagealign()))
			{
				sb.append("<td><img src=\"").append(page.getRoot()).append("images/").append(getImage()).append("\"");
				sb.append(" border=\"0\"></td>");
			}
			sb.append("</tr></table>");
		}
		sb.append("</td>\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}


}

