/*
	$Id: ExtraValueJoin.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;

import net.sf.basedb.core.query.Join;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.JoinType;

/**
	Represents a join to an extra value in a dynamic query:
	<code>leftTable JOIN extraValueTable ON ...</code>
	<p>
	This class is used by, for example, {@link 
	DynamicSpotQuery#joinExtraValue(ExtraValue, JoinType)} to create the 
	{@link Join} query element that is needed. After the extra value
	has been join it can be used in expressions in the query.


	@author Nicklas
	@version 2.0
	@see ExtraValueExpression
	@see net.sf.basedb.core.query.Dynamic#extraValue(ExtraValue)
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
class ExtraValueJoin
	implements Join
{
	
	private final ExtraValue extraValue;
	private final VirtualDb vdb;
	private final JoinType joinType;
	private final VirtualTable left;
	private final VirtualTable right;
	private final VirtualColumn[] columns;
	
	/**
		Create a new join using the root entity table as the left table.
		<code><i>SELECT ... FROM rootTable</i> JOIN extraValueTable</code>.
		<p>
		The extra value table and which columns should be used in the join
		is found by examining the extra value properties, ie. value type
		and coordinate type. 

		@param extraValue The extra value to join
		@param joinType The type of join
	*/
	ExtraValueJoin(ExtraValue extraValue, JoinType joinType)
	{
		this.extraValue = extraValue;
		this.vdb = extraValue.getVirtualDb();
		this.joinType = joinType;
		this.left = null;
		this.right = extraValue.getVirtualTable();
		ExtraValue.CoordinateType ct = extraValue.getCoordinateType();
		if (ct == ExtraValue.CoordinateType.SPOT)
		{
			this.columns = new VirtualColumn[] { VirtualColumn.CUBE, VirtualColumn.COLUMN, VirtualColumn.POSITION };
		}
		else if (ct == ExtraValue.CoordinateType.POSITION)
		{
			this.columns = new VirtualColumn[] { VirtualColumn.CUBE, VirtualColumn.POSITION };
		}
		else
		{
			this.columns = new VirtualColumn[] { VirtualColumn.CUBE, VirtualColumn.COLUMN };
		}
	}
	
	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : left.getAlias();
			String rightAlias = extraValue.getJoinAlias();
			
			StringBuilder sb = new StringBuilder();
			sb.append(joinType.getJoinExpression()).append(" ").append(right.getQualifiedTableName(vdb));
			sb.append(" AS ").append(rightAlias);
			sb.append(" ON ");
			sb.append(rightAlias).append(".").append(HibernateUtil.quote(VirtualColumn.EXTRA.getColumn()));
			sb.append(" = ").append(extraValue.getExtraNo());
			for (VirtualColumn column : columns)
			{
				String qualifiedColumn = HibernateUtil.quote(column.getColumn());
				sb.append(" AND ").append(rightAlias).append(".").append(qualifiedColumn);
				sb.append(" = ").append(leftAlias).append(".").append(qualifiedColumn);
			}
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ExtraValueJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			StringBuilder sb = new StringBuilder();
			sb.append(", ").append(right.getQualifiedTableName(vdb));
			sb.append(" ").append(HibernateUtil.quote(extraValue.getJoinAlias()));
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}

	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : HibernateUtil.quote(left.getAlias());
			String rightAlias = HibernateUtil.quote(extraValue.getJoinAlias());
			StringBuilder sb = new StringBuilder();
			String deliminator = "";
			for (VirtualColumn column : columns)
			{
				String qualifiedColumn = HibernateUtil.quote(column.getColumn());
				sb.append(deliminator);
				sb.append(leftAlias).append(".").append(qualifiedColumn);
				sb.append(" = ").append(rightAlias).append(".").append(qualifiedColumn);
				if (joinType == JoinType.LEFT) sb.append("(+)");
				deliminator = " AND ";
			}
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		String leftAlias = left == null ? "" : left.getAlias() + ".";
		String rightAlias = extraValue.getJoinAlias();

		StringBuilder sb = new StringBuilder();
		sb.append(joinType.getJoinExpression()).append(" ").append(right.getTableName(vdb));
		sb.append(" AS ").append(rightAlias);
		sb.append(" ON ");
		sb.append(rightAlias).append(".").append(VirtualColumn.EXTRA.getName());
		sb.append(" = ").append(extraValue.getExtraNo());
		for (VirtualColumn column : columns)
		{
			sb.append(" AND ").append(rightAlias).append(".").append(column.getName());
			sb.append(" = ").append(leftAlias).append(column.getName());
		}
		return sb.toString();
	}
	// -------------------------------------------

	
}
