/*
	$Id: ExtraValueType.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.Formula.AverageMethod;
import net.sf.basedb.core.data.ExtraValueTypeData;


/**
	Defines extra values that can be calculated by plugins or some other
	application and inserted as part of a <code>BioAssaySet</code>. All
	extra value types must have a unique external id that is used for 
	identification.
	<p>
	Currently only extra values of {@link Type#INT}, {@link Type#FLOAT}
	and {@link Type#STRING} are supported.

	@author Nicklas
	@version 2.0
	@see ExtraValue
	@see BioAssaySet#getSpotExtraValueBatcher(Class, ExtraValueType, Job)
	@see BioAssaySet#getPositionExtraValueBatcher(Class, ExtraValueType, Job)
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public class ExtraValueType
	extends BasicItem<ExtraValueTypeData>
	implements Nameable, Removable
{

	/**
		The type of item represented by this class.
		@see Item#EXTRAVALUETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.EXTRAVALUETYPE;

	/**
		The maximum length of the external id variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setExternalId(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = ExtraValueTypeData.MAX_EXTERNAL_ID_LENGTH;

	/**
		Create a new <code>ExtraValue</code> item. Currently only extra values 
		of {@link Type#INT}, {@link Type#FLOAT} and {@link Type#STRING} are supported.


		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param externalId A unique string identifying the extra value type
		@param valueType The <code>Type</code> of the extra value
		@return The new <code>ExtraValue</code> item
		@throws BaseException If there is an error
	*/
	public static ExtraValueType getNew(DbControl dc, String externalId, Type valueType)
		throws BaseException
	{
		ExtraValueType ev = dc.newItem(ExtraValueType.class);
		ev.setName("New extra value");
		ev.setExternalId(externalId);
		ev.setValueType(valueType);
		ev.setAverageMethod(valueType.isNumerical() ? AverageMethod.ARITHMETIC_MEAN : AverageMethod.NONE);
		return ev;
	}
	
	/**
		Get an <code>ExtraValueType</code> object when you know the id.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>ExtraValueType</code> item
		@throws ItemNotFoundException If an item with the specified id is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is an error
	*/
	public static ExtraValueType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ExtraValueType ev = dc.loadItem(ExtraValueType.class, id);
		if (ev == null) throw new ItemNotFoundException("ExtraValueType[id="+id+"]");
		return ev;
	}
	
	/**
		Get a <code>ExtraValueType</code> when you know the external id. 
		
		@param dc The <code>DbControl</code> which will be used for
			database access.
		@param externalId The external id of the extra value type to load
		@return The <code>ExtraValueType</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static ExtraValueType getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Session session = dc.getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_EXTRAVALUETYPE_FOR_EXTERNAL_ID");
		query.setString("externalId", externalId);
		ExtraValueType evt = dc.getItem(ExtraValueType.class, HibernateUtil.loadData(ExtraValueTypeData.class, query));
		if (evt == null) throw new ItemNotFoundException("ExtraValueType[externalId="+externalId+"]");
		return evt;
	}

	/**
		Get a query configured to retrieve extra value types.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ExtraValueType> getQuery()
	{
		return new ItemQuery<ExtraValueType>(ExtraValueType.class);
	}

	/**
		Creates a new extra value type item.
		@param data The data
	*/
	ExtraValueType(ExtraValueTypeData data)
	{
		super(data);
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Any {@link ExtraValue}:s of this type exists
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXTRAVALUES_FOR_EXTRAVALUETYPE", "count(*)");
		/*
			SELECT {1}
			FROM ExtraValueData ev
			WHERE ev.extraValueType = :extraValueType
		*/
		query.setEntity("extraValueType", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link ExtraValue}:s of this type
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXTRAVALUES_FOR_EXTRAVALUETYPE", "ev.id");
			/*
				SELECT {1}
				FROM ExtraValueData ev
				WHERE ev.extraValueType = :extraValueType
			*/
		query.setEntity("extraValueType", this.getData());
		addUsingItems(using, Item.EXTRAVALUE, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the value type for this extra value type. It can't be change once
		the object has been created.
	*/
	public Type getValueType()
	{
		return Type.fromValue(getData().getValueType());
	}

	/**
		Set the value type for this extra value type. Null is not
		allowed and it must be one of the supported value types.
		@see VirtualColumn#extraValue(Type)
	*/
	private void setValueType(Type valueType)
		throws InvalidDataException
	{
		if (valueType == null) throw new InvalidUseOfNullException("valueType");
		if (VirtualColumn.extraValue(valueType) == null)
		{
			throw new InvalidDataException("Value type is not supported as extra value: " + valueType);
		}
		getData().setValueType(valueType.getValue());
	}
	
	/**
		Get the external id for this <code>ExtraValueType</code> item.
		@return A string with the external id of this item
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}

	/**
		Set the external id for this <code>Client</code> item. The value
		cannot be null and must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant.
		@param externalId The new external id for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is null or longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNotNullString(externalId, "externalId", 
			MAX_EXTERNAL_ID_LENGTH));
	}

	private Coloring coloring;
	/**
		Get coloring information for this extra value.
		@return A {@link Coloring} object
	*/
	public Coloring getColoring()
	{
		if (coloring == null)
		{
			coloring = new Coloring(getData().getColoring(), this);
		}
		return coloring;
	}
	
	/**
		Get the average method for this extra value type.
		@return A {@link AverageMethod} object
		@since 2.4
	 */
	public Formula.AverageMethod getAverageMethod()
	{
		return AverageMethod.fromValue(getData().getAverageMethod());
	}
	/**
		Set the average method of the extra value type..
		@param averageMethod A {@link AverageMethod} object, null is not allowed
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the averageMethod is null
		@since 2.4
	*/
	public void setAverageMethod(Formula.AverageMethod averageMethod)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (averageMethod == null) throw new InvalidUseOfNullException("averageMethod");
		getData().setAverageMethod(averageMethod.getValue());
	}

}
