/*
	$Id: IntegerParameterType.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.IntegerParameterValueData;
import java.util.List;

/**
	This class represent a parameter type that is an integer.
	This type defines a {@link #getLowerLimit() lower limit}
	and an {@link #getUpperLimit upper limit} for the range of
	allowed values.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public class IntegerParameterType
	extends ParameterType<Integer>
{
	/**
		The lower limit of the parameter.
	*/
	private final Integer lowerLimit;

	/**
		The upper limit of the parameter.
	*/
	private final Integer upperLimit;


	/**
		Create a new integer parameter type, without any limits and allowing null
		values.
	*/
	public IntegerParameterType()
	{
		this(null, null, null, false);
	}
	
	/**
		Create a new integer parameter type.
		@param lowerLimit The lowest allowed value, or null to have no limit
		@param upperLimit The highest allowed value, or null to have no limit
		@param defaultValue The default value for this parameter
		@param notNull FALSE if nulls values are allowed, TRUE otherwise
	*/
	public IntegerParameterType(Integer lowerLimit, Integer upperLimit, Integer defaultValue, boolean notNull)
	{
		super(Integer.class, defaultValue, notNull);
		this.lowerLimit = lowerLimit;
		this.upperLimit = upperLimit;
	}

	public IntegerParameterType(Integer lowerLimit, Integer upperLimit, Integer defaultValue, boolean notNull, 
		int multiplicity, int width, int height, List<Integer> items)
	{
		super(Integer.class, defaultValue, notNull, multiplicity, width, height, items);
		this.lowerLimit = lowerLimit;
		this.upperLimit = upperLimit;
	}
	
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		return "IntegerParameterType: Lower " + lowerLimit + ", Upper " + upperLimit;
	}
	// -------------------------------------------
	
	/*
		From the ParameterType class
		-------------------------------------------
	*/
	/**
		Checks if the value is within the range given by the upper and lower 
		limits if those are given.
		@param value The value to test
		@throws InvalidDataException If the value is outside the range of allowed values
	*/
	@Override
	void validateValue(String name, Integer value)
		throws InvalidDataException
	{
		if (((lowerLimit != null) && (value.compareTo(lowerLimit) < 0 ))
			||
			((upperLimit != null) && (value.compareTo(upperLimit) > 0 )))
		{
			if (lowerLimit != null && upperLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, lowerLimit, upperLimit);
			}
			else if (lowerLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, lowerLimit, false);
			}
			else if (upperLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, upperLimit, true);
			}

		}
	}
	/**
		Create a new {@link IntegerParameterValueData} object.
	*/
	@Override
	IntegerParameterValueData newParameterValueData()
	{
		return new IntegerParameterValueData();
	}
	// -------------------------------------------

	/**
		Get the lowest valid value of the parameter.
		@return Lowest valid value or null if none is set.
	*/
	public Integer getLowerLimit()
	{
		return lowerLimit;
	}
	
	/**
		Get the highest valid value of the parameter.
		@return Highest valid value or null if none is set.
	*/
	public Integer getUpperLimit()
	{
		return upperLimit;
	}
}
