/*
	$Id: PluginConfiguration.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.plugin.AnnotationSetterPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.data.PluginConfigurationData;
import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.data.VersionedParameter;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Expressions;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Collections;

/**
	This class is used to configure a plugin. A plugin can have many
	configurations. The sequence to setup a configuration is the following

	<ul>
		<li> TODO Description on how to configure
	</ul>

	@author Nicklas, Samuel
	@version 2.0
*/
public class PluginConfiguration
	extends AnnotatedItem<PluginConfigurationData>
	implements Transactional, AnnotatableProxy
{
	/**
		The type of item represented by this class.
		@see Item#PLUGINCONFIGURATION
		@see #getType()
	*/
	public static final Item TYPE = Item.PLUGINCONFIGURATION;


	/**
		Create a new <code>PluginConfiguration</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
	 	@param pd PluginDefinition to create a configuration for. 
		@return The new <code>PluginConfiguration</code> item
		@throws InvalidDataException If the plugin definition is null
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for the plugin definition
		@throws BaseException If there is an error
	*/
	public static PluginConfiguration getNew(DbControl dc, PluginDefinition pd)
		throws PermissionDeniedException, BaseException
	{
		PluginConfiguration pc = dc.newItem(PluginConfiguration.class);
		pc.setPluginDefinition(pd);
		return pc;
	}


	/**
		Get a <code>PluginConfiguration</code> item when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>PluginConfiguration</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ READ} permission for the item
		@throws BaseException If there is another error
	*/
	public static PluginConfiguration getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PluginConfiguration p = dc.loadItem(PluginConfiguration.class, id);
		if (p == null) throw new ItemNotFoundException("PluginConfiguration[id="+id+"]");
		return p;
	}


	/**
		Get a query configured to retrieve plugin configurations.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PluginConfiguration> getQuery()
	{
		return new ItemQuery<PluginConfiguration>(PluginConfiguration.class);
	}

	/**
		Get a query that returns all plugins configurations related
		to a specified GUI context and implementing a certain interface. 
		Ie. The {@link PluginDefinition#isInContext(GuiContext)} returns true for the 
		specified item, and {@link PluginDefinition#supports(PluginType)} returns 
		true for the specified interface.
		
		@param gc The item the plugin should be related to, or null
			if this parameter is irrelevant
		@param interfaceName The complete name of the interface that the plugin
			must implement, or null if this parameter is irrelevant
	*/
	public static ItemQuery<PluginConfiguration> getQuery(GuiContext gc, String interfaceName)
	{
		ItemQuery<PluginConfiguration> query = getQuery();
		Item item = gc == null ? null : gc.getItem();
		GuiContext.Type type = gc == null ? null : gc.getType();
		if (item != null || type != null)
		{
			query.joinPermanent(Hql.innerJoin("pluginDefinition.guiContexts", "gcx"));
			if (item != null)
			{
				query.restrictPermanent(
					Restrictions.eq(
						Hql.property("gcx", "itemType"), 
						Expressions.integer(item.getValue())
					)
				);
			}
			if (type != null)
			{
				query.restrictPermanent(
					Restrictions.eq(
						Hql.property("gcx", "contextType"), 
						Expressions.integer(type.getValue())
					)
				);
			}
		}
		if (interfaceName != null)
		{
			query.joinPermanent(Hql.innerJoin("pluginDefinition.pluginTypes", Item.PLUGINTYPE.getAlias()));
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property(Item.PLUGINTYPE.getAlias(), "interfaceName"), 
					Expressions.parameter("interfaceName")
				)
			);
			query.setPermanentParameter("interfaceName", interfaceName, Type.STRING);
		}
		return query;
	}

	PluginConfiguration(PluginConfigurationData pc)
	{
		super(pc);
	}


	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A {@link Job} is using this plugin configuration
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_JOBS_FOR_PLUGINCONFIGURATION", "count(*)");
		/*
			SELECT {1}
			FROM JobData j
			WHERE j.pluginConfiguration = :pluginConfiguration
		*/
		query.setEntity("pluginConfiguration", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Job}:s using this configuration
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_JOBS_FOR_PLUGINCONFIGURATION", "j.id");
			/*
				SELECT {1}
				FROM JobData j
				WHERE j.pluginConfiguration = :pluginConfiguration
			*/
		query.setEntity("pluginConfiguration", this.getData());
		addUsingItems(using, Item.JOB, query);
		return using;
	}
	/**
		Reset parameter version so that adding new parameters will
		increase the version the next time.
	*/
	void onAfterCommit(Transactional.Action action)
	{
		newParameterVersion = 0;
	}
	// -------------------------------------------
	/*
		From the Annotatable & AnnotatableProxy interfaces
		-------------------------------------------
	*/
	/**
		@return Always null
		@since 2.5
	*/
	public Set<Annotatable> getAnnotatableParents()
	{
		return null;
	}
	/**
		Get the annotation types that can be used by this configuration. This method
		requires that the plug-in implements the {@link AnnotationSetterPlugin} interface.
		@return A query or null if this plug-in doesn't use annotations
		@since 2.5
	*/
	public ItemQuery<AnnotationType> getAnnotationTypes()
	{
		Plugin p = getPluginDefinition().newInstance(Plugin.class, null, getSessionControl(), this, null);
		ItemQuery<AnnotationType> query = null;
		if (p instanceof AnnotationSetterPlugin)
		{
			query = ((AnnotationSetterPlugin)p).getAnnotationTypes();
		}
		return query;
	}
	/**
		Get a message explaining what the proxy does with the annotations.
		@return A message or null
		@since 2.5
	*/
	public String getAnnotationMessage()
	{
		Plugin p = getPluginDefinition().newInstance(Plugin.class, null, getSessionControl(), this, null);
		String message = null;
		if (p instanceof AnnotationSetterPlugin)
		{
			message = ((AnnotationSetterPlugin)p).getAnnotationMessage();
		}
		return message;
	}
	// -------------------------------------------

	/**
		Get the plugin definition for this configuration.
		@return <code>PluginDefinition</code> object.
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#READ READ}
			permission to the items.
		@throws BaseException If anything else fails.
	*/
	public PluginDefinition getPluginDefinition()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PluginDefinition.class, getData().getPluginDefinition());
	}
	/**
		The definition cannot be changed. Used internally only.
	*/
	private void setPluginDefinition(PluginDefinition pd)
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		if (pd == null) throw new InvalidUseOfNullException("pd");
		pd.checkPermission(Permission.USE);
		getData().setPluginDefinition(pd.getData());
	}

	/**
		Get the latest version number of the parameters.
	*/
	public int getParameterVersion()
	{
		return getData().getParameterVersion();
	}
	
	private int newParameterVersion = 0;
	/**
		Increase the version number once per transaction.
	*/
	private int getNewParameterVersion()
	{
		if (newParameterVersion == 0)
		{
			newParameterVersion = getParameterVersion() + 1;
			getData().setParameterVersion(newParameterVersion);
		}
		return newParameterVersion;
	}

	/**
		Get the names of all configuration values in the latest version.
		@return A <code>Set</code> containing the names of the configuration values
			or an empty set if no values has been defined
		@see #getParameterNames(int)
	*/
	public Set<String> getParameterNames()
	{
		return getParameterNames(getParameterVersion());
	}
	
	/**
		Get the names of all configuration values in a specific version.
		@param version The version to get the configuration parameters for
		@return A <code>Set</code> containing the names of the configuration values
			or an empty set if no values has been defined
		@see #getParameterNames()
	*/
	public Set<String> getParameterNames(int version)
	{
		Set<String> names = new HashSet<String>();
		for (VersionedParameter param : getData().getConfigurationValues().keySet())
		{
			if (param.getVersion() == version) names.add(param.getName());
		}
		return Collections.unmodifiableSet(names);
	}
	
	/**
		Get the values of a configuration parameter in the latest version.
		@param name The name of the parameter
		@return The values, or null if no value has been set
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to all values
		@throws BaseException If there is another error
		@see #getParameterValues(String, int)
	*/
	public List<?> getParameterValues(String name)
		throws PermissionDeniedException, BaseException
	{
		return getParameterValues(name, getParameterVersion());
	}

	/**
		Get the values of a configuration parameter in a specific version.
		@param name The name of the parameter
		@param version The version of the parameter
		@return The values, or null if no value has been set
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to all values
		@throws BaseException If there is another error
		@see #getParameterValues(String)
	*/
	public List<?> getParameterValues(String name, int version)
		throws PermissionDeniedException, BaseException
	{
		VersionedParameter param = new VersionedParameter(name, version);
		ParameterValueData<?> value = getData().getConfigurationValues().get(param);
		if (value == null) return null;
		return Collections.unmodifiableList(Values.getItemValues(getDbControl(), value.getValues()));
	}

	
	/**
		Get extra information about a parameter in the latest version. The information 
		includes the label, description and values of the parameter.
		@param name The name of the parameter
		@return A ParameterInfo object or null if no parameter with the 
			specified name exists
		@throws PermissionDeniedException If the parameter contain values that
			the logged in user doesn't have read permission for
		@throws BaseException If there is another error
		@see #getParameterInfo(String, int)
	*/
	public ParameterInfo getParameterInfo(String name)
		throws PermissionDeniedException, BaseException
	{
		return getParameterInfo(name, getParameterVersion());
	}
	
	/**
		Get extra information about a parameter in a specific version. The information 
		includes the label, description and values of the parameter.
		@param name The name of the parameter
		@param version The version of the parameter
		@return A ParameterInfo object or null if no parameter with the 
			specified name exists
		@throws PermissionDeniedException If the parameter contain values that
			the logged in user doesn't have read permission for
		@throws BaseException If there is another error
		@see #getParameterInfo(String)
	*/
	public ParameterInfo getParameterInfo(String name, int version)
		throws PermissionDeniedException, BaseException
	{
		VersionedParameter param = new VersionedParameter(name, version);
		ParameterValueData<?> parameterValue = getData().getConfigurationValues().get(param);
		if (parameterValue == null) return null;
		return new ParameterInfo(getDbControl(), parameterValue);
	}

	/**
		Get the configuration parameters as a {@link ParameterValues} object.
		@param version The version of the parameters
		@return A ParameterValues object
		@see #getParameterVersion()
		@since 2.5
	*/
	public ParameterValues getParameterValues(int version)
	{
		return getParameterValuesImpl(getParameterVersion());
	}
	
	/**
		Set a configuration parameter. The parameter version number is increased
		by one when the values are committed to the database.

		@param name The name of the configuration parameter
		@param parameterType The type of the parameter
		@param values A list containing the new values, null or empty to remove the 
			configuration values
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValues(String name, ParameterType<?> parameterType, List<?> values)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		setParameterValues(name, null, null, parameterType, values);
	}

	/**
		Set a configuration parameter. The parameter version number is increased
		by one when the values are committed to the database.
	
		@param name The name of the configuration parameter
		@param label The label of the parameter (optional)
		@param description A description of the parameter (optional)
		@param parameterType The type of the parameter
		@param values A list containing the new values, null or empty to remove the 
			configuration values
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValues(String name, String label, String description, ParameterType<?> parameterType, List<?> values)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (name == null) throw new InvalidUseOfNullException("name");
		setParameterValuesInternal(name, label, description, parameterType, values, true);
	}

	/**
		Set the values of a configuration parameter.
		@param name The name of the configuration parameter
		@param label The label of the parameter (optional)
		@param description A description of the parameter (optional)
		@param parameterType The type of the parameter
		@param values A list containing the new values, null or empty to remove the 
			configuration values
		@param validate If validation by {@link ParameterType#validate(String, List)}
			is needed or not
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	void setParameterValuesInternal(String name, String label, String description, 
		ParameterType<?> parameterType, List<?> values, boolean validate)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{

		VersionedParameter param = new VersionedParameter(name, getNewParameterVersion());
		if (values == null || values.size() == 0)
		{
			getData().getConfigurationValues().remove(param);
		}
		else
		{
			if (parameterType == null) throw new InvalidUseOfNullException("parameterType");
			if (validate) parameterType.validate(name, values);
			ParameterValueData<?> parameterValue = parameterType.newParameterValueData();
			parameterValue.setLabel(label);
			parameterValue.setDescription(description);
			parameterValue.replaceValues(Values.getDataValues(values));
			getData().getConfigurationValues().put(param, parameterValue);
		}
	}
	
	/**
		Copy all parameter values from another plugin configuration. This method
		increases the parameter version number and saves the copied parameters
		in that version.
		@param copyFrom The configuration to copy parameters from
		@throws InvalidDataException If the copyFrom parameter is null
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission for this configuration
		@throws BaseException If there is another error
	*/
	@SuppressWarnings("unchecked")
	public void copyParametersFrom(PluginConfiguration copyFrom)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (copyFrom == null) throw new InvalidUseOfNullException("copyFrom");
		Map<VersionedParameter, ParameterValueData<?>> to = getData().getConfigurationValues();
		Map<VersionedParameter, ParameterValueData<?>> from = copyFrom.getData().getConfigurationValues();

		int newVersion = getNewParameterVersion();
		int copyFromVersion = copyFrom.getParameterVersion();
		try
		{
			for (Map.Entry<VersionedParameter, ParameterValueData<?>> fromEntry : from.entrySet())
			{
				VersionedParameter fromName = fromEntry.getKey();
				if (copyFromVersion == fromName.getVersion())
				{
					VersionedParameter toName = new VersionedParameter(fromName.getName(), newVersion);
					ParameterValueData<?> fromParameter = fromEntry.getValue();
					Class<? extends ParameterValueData> dataClass = fromParameter.getClass();
					if (org.hibernate.proxy.HibernateProxy.class.isAssignableFrom(dataClass))
					{
						// The 'fromParameter' is a proxy, we need the real data class
						dataClass = 
							(Class<? extends ParameterValueData>)org.hibernate.Hibernate.getClass(fromParameter);
					}
					ParameterValueData<?> toParameter = dataClass.newInstance();
					toParameter.setLabel(fromParameter.getLabel());
					toParameter.setDescription(fromParameter.getDescription());
					toParameter.replaceValues(fromParameter.getValues());
					to.put(toName, toParameter);
				}
			}
		}
		catch (IllegalAccessException ex)
		{
			throw new BaseException(ex);
		}
		catch (InstantiationException ex)
		{
			throw new BaseException(ex);
		}
		
	}
	
	/**
		Start the configuration sequence for a plugin. See {@link net.sf.basedb.core.PluginRequest} 
		for more  information. The parameters sent to the plugin are those from the latest
		version.
		
		@return A <code>PluginRequest</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws BaseException If there is another error
	*/
	public PluginConfigurationRequest configure()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		PluginDefinition pd = getPluginDefinition();
		if (!pd.isInteractive())
		{
			throw new BaseException("Plugin can't be configured interactively: " + pd);
		}
		InteractivePlugin plugin = (InteractivePlugin)pd.newInstance(); 
		SessionControl sc = getSessionControl();
		ParameterValuesImpl parameters = new ParameterValuesImpl(this, getParameterVersion(), false);
		
		PluginConfigurationRequest request = 
			new PluginConfigurationRequest(
				sc, plugin, Request.COMMAND_CONFIGURE_PLUGIN,
				this, parameters, null, null, null, pd, false
			);
		return request;
	}
	
	/**
		Create a new job.
		@deprecated Use {@link #newJob(Experiment)} instead
	*/
	public Job newJob()
		throws PermissionDeniedException, BaseException
	{
		Job j = Job.getNew(getDbControl(), getPluginDefinition(), this, null);
		return j;
	}
	
	/**
		Create a new {@link Job} for this plugin configuration.
		@param experiment The experiment the job is part of, or null
			if the job is not part of an experiment
		@return The new <code>Job</code> item
		@throws PermissionDeniedException If
			the logged in user doesn't have use
			permission for this plugin definition
		@since 2.7
	*/
	public Job newJob(Experiment experiment)
		throws PermissionDeniedException, BaseException
	{
		Job j = Job.getNew(getDbControl(), getPluginDefinition(), this, experiment);
		return j;
	}
	/**
		Get a query that returns the jobs 
		using this plugin configuration.

		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Job> getJobs()
	{
		ItemQuery<Job> query = Job.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("pluginConfiguration"), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a write-protected ParameterValuesImpl object so a job can read
		the paremeters.
	*/
	ParameterValuesImpl getParameterValuesImpl(int version)
		throws BaseException
	{
		return new ParameterValuesImpl(this, version, true);
	}
	

}
