/*
	$Id: QueryRuntimeFilterFactory.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.query.EntityQuery;

import java.util.List;
import java.util.LinkedList;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.Arrays;

/**
	This is a factory class for {@link QueryRuntimeFilter} implementations.
	The factory creates required and default optional filters for all items
	at startup time.

	@author Nicklas
	@version 2.0
	@see <a href="../../../../../development/overview/query/overview.html#factory">Query API overview - filter factory</a>
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
class QueryRuntimeFilterFactory
{

	private static Map<Item, QueryRuntimeFilter> requiredFilters;
	private static Map<Item, QueryRuntimeFilter> optionalFilters;
	public static final Set<Integer> ZERO_SET = Collections.singleton(0);

	static final QueryRuntimeFilter DENY_ALL_FILTER = new DenyAllFilter();
	static final QueryRuntimeFilter REMOVABLE_FILTER = new RemovableFilter();
	static final QueryRuntimeFilter ANNOTATABLE_FILTER = new AnnotatableFilter();
	static final QueryRuntimeFilter BASIC_FILTER = new BasicFilter();
	static final QueryRuntimeFilter OWNABLE_FILTER = new OwnableFilter();
	static final QueryRuntimeFilter SHAREABLE_FILTER = new ShareableFilter();

	private static boolean isInitialised = false;

	static synchronized void init()
	{
		if (isInitialised) return;
		requiredFilters = new HashMap<Item, QueryRuntimeFilter>();
		optionalFilters = new HashMap<Item, QueryRuntimeFilter>();
		for (Item item : Item.values())
		{
			Class c = item.getDataClass();
			List<QueryRuntimeFilter> requiredChain = new LinkedList<QueryRuntimeFilter>();
			List<QueryRuntimeFilter> optionalChain = new LinkedList<QueryRuntimeFilter>();
			requiredChain.add(DENY_ALL_FILTER);
			if (c != null)
			{
				if (RemovableData.class.isAssignableFrom(c))
				{
					requiredChain.add(REMOVABLE_FILTER);
				}
				if (AnnotatableData.class.isAssignableFrom(c))
				{
					requiredChain.add(ANNOTATABLE_FILTER);
				}
				if (ShareableData.class.isAssignableFrom(c))
				{
					optionalChain.add(SHAREABLE_FILTER);
				}
				else if (OwnableData.class.isAssignableFrom(c))
				{
					optionalChain.add(OWNABLE_FILTER);
				}
				else
				{
					optionalChain.add(BASIC_FILTER);
				}
				requiredFilters.put(item, new ChainedFilter(requiredChain));
				optionalFilters.put(item, new ChainedFilter(optionalChain));
			}
		}
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (requiredFilters != null) requiredFilters.clear();
		requiredFilters = null;
		if (optionalFilters != null) optionalFilters.clear();
		optionalFilters = null;
	}

	/**
		Get the required runtime filter for the speicified item type.
		@param itemType The item type
		@return A <code>QueryRuntimeFilter</code> object or null
			if not filter is required
	*/
	static QueryRuntimeFilter getRequiredFilter(Item itemType)
	{
		return requiredFilters.get(itemType);
	}
	
	/**
		Get the default optional runtime filter for the speicified item type.
		@param itemType The item type
		@return A <code>QueryRuntimeFilter</code> object or null
			if not filter is required
	*/
	static QueryRuntimeFilter getOptionalFilter(Item itemType)
	{
		return optionalFilters.get(itemType);
	}


	/**
		A required filter for all items, which checks if the logged in user
		has been "granted" the {@link Permission#DENIED} permission for
		a specific item type. If that is the case the <code>denyAll</code> filter
		is enabled.
	*/
	private static class DenyAllFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			if (dc.getSessionControl().hasPermission(Permission.DENIED, query.getItemType()))
			{
				// No items should be returned if a user is denied access
				manager.enableFilter("denyAll");
			}
		}
	}

	/**
		A required filter for {@link Removable} items. The filter checks the query
		for the {@link Include#REMOVED} and {@link Include#NOT_REMOVED} options
		and enabled the <code>isRemoved</code> filter if neccessary.
	*/
	private static class RemovableFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			if (!query.isIncluded(Include.REMOVED, Include.NOT_REMOVED))
			{
				// We do not want BOTH "removed" and "not removed" items
				org.hibernate.Filter filter = manager.enableFilter("isRemoved");
				if (filter != null)
				{
					filter.setParameter("removed", query.isIncluded(Include.REMOVED));
				}
			}
		}
	}

	/**
		A required filter for {@link Annotatable} items. The filter checks the query
		for the {@link Include#ANNOTATED} and {@link Include#NOT_ANNOTATED} options
		and enabled the <code>isAnnotated</code> filter if neccessary. If none of
		the options are set this is the same as if both options are set.
		@since 2.5
	*/
	private static class AnnotatableFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			boolean annotated = query.isIncluded(Include.ANNOTATED);
			boolean notAnnotated = query.isIncluded(Include.NOT_ANNOTATED);
			if (annotated ^ notAnnotated)
			{
				// Only one option is set (both or none requires no filter)
				org.hibernate.Filter filter = manager.enableFilter("isAnnotated");
				if (filter != null)
				{
					filter.setParameter("annotated", annotated);
				}
			}
		}
	}
	
	/**
		An optional default filter for {@link BasicItem}:s that are not
		{@link Ownable} or {@link Shareable}. The filter checks if the logged in user
		has generic permission {@link EntityQuery#getItemPermission()} for a specific item type. 
		If not, the <code>denyAll</code> filter is enabled. This filter is often replaced by 
		another filter for example the {@link ChildFilter}.
	*/
	private static class BasicFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			if (!dc.getSessionControl().hasPermission(query.getItemPermission(), query.getItemType()))
			{
				// No items should be returned if a user doesn't have access
				manager.enableFilter("denyAll");
			}
		}
	}

	/**
		An optional filter that can be used by {@link BasicItem}:s to replace
		the {@link BasicFilter}. This filter checks if the logged in user has
		generic permission {@link EntityQuery#getItemPermission()} for a parent item,
		instead of the actual item returned by a query. In a parent/child relationship it is the
		permissions on the parent item that controls the permission on the child.
		For example, {@link Client} is the parent of {@link Help}. 
		<p>
		If the logged in user doesn't have the requested permission on the parent
		item, the <code>denyAll</code> filter is enabled.
	*/
	static class ChildFilter
		implements QueryRuntimeFilter
	{
		/**
			The item type of the child item.
		*/
		private final Item child;
		/**
			The item type of the parent item.
		*/
		private final Item parent;
		
		/**
			Create a new <code>ChildFilter</code> object.
			@param child The item type of the child item
			@param parent The item type of the parent item
		*/
		ChildFilter(Item child, Item parent)
		{
			this.child = child;
			this.parent = parent;
		}
	
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			if (query.getItemType() != child && dc.getSessionControl().isDenied(query.getItemPermission(), parent))
			{
				// No items should be returned if a user is denied access
				manager.enableFilter("denyAll");
			}
		}
	}

	/**
		An optional default filter for {@link Ownable} items that are not
		{@link Shareable}. The filter checks the query for the {@link Include#MINE} and
		{@link Include#OTHERS} options and enabled one of the <code>ownedBy</code>,
		<code>notOwnedBy</code> and <code>denyAll</code> filters if neccessary.
		If the logged in user doesn't have generic permission {@link EntityQuery#getItemPermission()} 
		for a specific item type, items owned by other users will not be returned.
	*/
	private static class OwnableFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			org.hibernate.Filter filter = null;
			SessionControl sc = dc.getSessionControl();
			int owner = sc.getLoggedInUserId();
			
			if (sc.isDenied(query.getItemPermission(), query.getItemType()))
			{
				manager.enableFilter("denyAll");
			}
			else if (query.isIncluded(Include.OTHERS) && 
				sc.hasPermission(query.getItemPermission(), query.getItemType()))
			{
				// Logged in user has generic permission and wants other user's items...
				if (!query.isIncluded(Include.MINE))
				{
					// ... but not the one's with ownership
					filter = manager.enableFilter("notOwnedBy");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
					}
				}
			}
			else if (query.isIncluded(Include.MINE))
			{
				//Logged in user only wants own items
				filter = manager.enableFilter("ownedBy");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
				}
			}
			else
			{
				manager.enableFilter("denyAll");
			}
		}
	}

	/**
		An optional default filter for {@link Shareable} items. The filter checks
		the query for the {@link Include#MINE}, {@link Include#OTHERS},
		{@link Include#SHARED} and {@link Include#IN_PROJECT} and enables
		one of the filters:
		<ul>
		<li><code>ownedBy</code>
		<li><code>notOwnedBy</code>
		<li><code>sharedTo</code>
		<li><code>inProject</code>
		<li><code>ownedByOrSharedTo</code>
		<li><code>ownedByOrInProject</code>
		<li><code>sharedToOrInProject</code>
		<li><code>ownedByOrSharedToOrInProject</code>
		<li><code>denyAll</code>
		</ul>
		The filter makes sure that the query doesn't return an item that
		the logged in user doesn't have the permission {@link EntityQuery#getItemPermission()} 
		for.
	*/
	private static class ShareableFilter
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			org.hibernate.Filter filter = null;
			SessionControl sc = dc.getSessionControl();
			
			// The logged in user
			int owner = sc.getLoggedInUserId();
			
			// The required permission
			Permission permission = query.getItemPermission();
			
			if (sc.isDenied(permission, query.getItemType()))
			{
				manager.enableFilter("denyAll");
			}
			else
			{
				// The currently active project
				int activeProjectId = sc.getActiveProjectId();
				
				// All items keys for the logged in user
				Set<Integer> itemKeys = sc.getItemKeys(permission);
				if (itemKeys == null || itemKeys.size() == 0) itemKeys = ZERO_SET;
				
				// All project keys for the active project and logged in user
				Set<Integer> projectKeys = sc.getProjectKeys(permission);
				if (projectKeys == null || projectKeys.size() == 0) projectKeys = ZERO_SET;

				if (query.isIncluded(Include.OTHERS) && sc.hasPermission(permission, query.getItemType()))
				{
					// Logged in user has generic permission and wants other user's items...
					if (!query.isIncluded(Include.MINE))
					{
						// ... but not the one's with ownership
						filter = manager.enableFilter("notOwnedBy");
						if (filter != null)
						{
							filter.setParameter("owner", owner);
						}
					}
				}
				else if (query.isIncluded(Include.MINE, Include.SHARED, Include.IN_PROJECT) && activeProjectId != 0)
				{
					// Logged in user wants own, shared and project items
					filter = manager.enableFilter("ownedByOrSharedToOrInProject");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("itemKeys", itemKeys);
						filter.setParameterList("projectKeys", projectKeys);
					}
				}
				else if (query.isIncluded(Include.MINE, Include.IN_PROJECT) && activeProjectId != 0)
				{
					// Logged in user wants own and project items, this is the default query options
					filter = manager.enableFilter("ownedByOrInProject");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("projectKeys", projectKeys);
					}
				}
				else if (query.isIncluded(Include.SHARED, Include.IN_PROJECT) && activeProjectId != 0)
				{
					// Logged in user wants shared and project item, skipping the owned ones
					filter = manager.enableFilter("sharedToOrInProject");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("itemKeys", itemKeys);
						filter.setParameterList("projectKeys", projectKeys);
					}
				}
				else if (query.isIncluded(Include.MINE, Include.SHARED))
				{
					// Logged in user wants own and shared items
					filter = manager.enableFilter("ownedByOrSharedTo");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("itemKeys", itemKeys);
					}
				}
				else if (query.isIncluded(Include.IN_PROJECT) && activeProjectId != 0)
				{
					// Logged in user only wants project items
					filter = manager.enableFilter("inProject");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("projectKeys", projectKeys);
					}
				}
				else if (query.isIncluded(Include.SHARED))
				{
					// Logged in user only wants shared items
					filter = manager.enableFilter("sharedTo");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
						filter.setParameterList("itemKeys", itemKeys);
					}
				}
				else if (query.isIncluded(Include.MINE))
				{
					// Logged in user only wants own items, this is the default query options when no project is active
					filter = manager.enableFilter("ownedBy");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
					}
				}
				else
				{
					// Logged in user doesn't want any items at all
					manager.enableFilter("denyAll");
				}
			}
		}
	}
	
	/**
		Chains one or more filters into a single filter. This filter will simply
		call {@link QueryRuntimeFilter#enableFilters(QueryRuntimeFilterManager, EntityQuery, DbControl)}
		for each of the chained filters.
	*/
	static class ChainedFilter
		implements QueryRuntimeFilter
	{

		/**
			The list of chained filters.
		*/
		private final List<QueryRuntimeFilter> filterChain;
	
		/**
			Create a new <code>ChainedFilter</code> object.
		*/
		ChainedFilter(List<QueryRuntimeFilter> filterChain)
		{
			this.filterChain = filterChain;
		}
		
		/**
			Create a new <code>ChainedFilter</code> object.
		*/
		ChainedFilter(QueryRuntimeFilter... filterChain)
		{
			this.filterChain = Arrays.asList(filterChain);
		}

		/**
			Call {@link QueryRuntimeFilter#enableFilters(QueryRuntimeFilterManager, EntityQuery, DbControl)}
			for each of the chained filters.
		*/
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			for (QueryRuntimeFilter filter : filterChain)
			{
				filter.enableFilters(manager, query, dc);	
			}
		}
	}

}
