/*
  $Id: AnnotationTypeData.java 4544 2008-09-25 07:01:30Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;

/**
	This class holds information about an annotation type.
 
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.AnnotationType
	@see <a href="../../../../../../html/developerdoc/api/data_api.annotations.html">Annotations overview</a>
	@base.modified $Date: 2008-09-25 09:01:30 +0200 (Thu, 25 Sep 2008) $
	@hibernate.class table="`AnnotationTypes`" lazy="false"
*/
public class AnnotationTypeData
	extends CommonData
{

	public AnnotationTypeData()
	{}

	private QuantityData quantity;
	/**
		Get the quantity this annotation type uses. Can't be changed
		after creation.
		@hibernate.many-to-one column="`quantity_id`" not-null="false" outer-join="false"
		@since 2.9
	*/
	public QuantityData getQuantity()
	{
		return quantity;
	}
	/**
		@since 2.9
	*/
	public void setQuantity(QuantityData quantity)
	{
		this.quantity = quantity;
	}
	
	private UnitData defaultUnit;
	/**
		Get the unit used to store annotation values for this type. 
		NOTE! If the unit changes to another unit with a different 
		reference factor all annotation values must be recalculated.
		NOTE! This value is always null if the quantity is null.
		@hibernate.many-to-one column="`default_unit_id`" not-null="false" outer-join="false" 
		@since 2.9
	*/
	public UnitData getDefaultUnit()
	{
		return defaultUnit;
	}
	/**
		@since 2.9
	*/
	public void setDefaultUnit(UnitData defaultUnit)
	{
		this.defaultUnit = defaultUnit;
	}

	private Set<UnitData> usableUnits;
	/**
		The units that it makes sense to use with this annotation type.
		If empty, any unit for the given quantity can be used.
		@hibernate.set table="`AnnotationTypeUnits`" lazy="true" 
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-many-to-many column="`unit_id`" class="net.sf.basedb.core.data.UnitData"
		@since 2.9
	*/
	public Set<UnitData> getUsableUnits()
	{
		if (usableUnits == null) usableUnits = new HashSet<UnitData>();
		return usableUnits;
	}
	
	/**
		@since 2.9
	*/
	void setUsableUnits(Set<UnitData> usableUnits)
	{
		this.usableUnits = usableUnits;
	}
	
	/**
		The maximum length of the external id that can be stored in the database.
		@see #setExternalId(String)
		@since 2.2
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = 255;
	private String externalId;
	/**
		Get the external id for the biomaterial
		@hibernate.property column="`external_id`" type="string" length="255" not-null="false"
		@since 2.2
	*/
	public String getExternalId()
	{
		return externalId;
	}
	public void setExternalId(String externalId)
	{
		this.externalId = externalId;
	}
	
	private int multiplicity;
	/**
		How many values are allowed for this annotations of this type. 0 = no limit
		@hibernate.property column="`multiplicity`" type="int" not-null="true"
	*/
	public int getMultiplicity()
	{
		return multiplicity;
	}
	public void setMultiplicity(int multiplicity)
	{
		this.multiplicity = multiplicity;
	}

	private boolean requiredForMiame;
	/**
		If an annotation of this type must be present for an annotatable item to be
		compliant with MIAME.
		@hibernate.property column="`required_for_miame`" type="boolean" not-null="true"
	*/
	public boolean isRequiredForMiame()
	{
		return requiredForMiame;
	}
	public void setRequiredForMiame(boolean requiredForMiame)
	{
		this.requiredForMiame = requiredForMiame;
	}

	private boolean isEnumeration;
	/**
		If this annotation type defines a list of allowed values or not.
		@hibernate.property column="`is_enumeration`" type="boolean" not-null="true"
	*/
	public boolean isEnumeration()
	{
		return isEnumeration;
	}
	public void setEnumeration(boolean isEnumeration)
	{
		this.isEnumeration = isEnumeration;
	}
	
	private int height;
	/**
		The recommended height (lines) to display input fields for this 
		annotation in client applications.
		@hibernate.property column="`height`" type="int" not-null="true"
	*/
	public int getHeight()
	{
		return height;
	}
	public void setHeight(int height)
	{
		this.height = height;
	}

	private int width;
	/**
		The recommended width (characters) to display input fields for this 
		annotation in client applications.
		@hibernate.property column="`width`" type="int" not-null="true"
	*/
	public int getWidth()
	{
		return width;
	}
	public void setWidth(int width)
	{
		this.width = width;
	}

	private String defaultValue;
	public static final int MAX_DEFAULT_VALUE_LENGTH = 255;
	/**
		The default value of new annotations.
		@hibernate.property column="`default_value`" type="string" length="255" not-null="false"
	*/
	public String getDefaultValue()
	{
		return defaultValue;
	}
	public void setDefaultValue(String defaultValue)
	{
		this.defaultValue = defaultValue;
	}

	private int valueType;
	/**
		The code for the value type.
		@hibernate.property column="`value_type`" type="int" not-null="true" update="false"
	*/
	public int getValueType()
	{
		return valueType;
	}
	public void setValueType(int valueType)
	{
		this.valueType = valueType;
	}

	private Set<Integer> itemTypes;
	/**
		The item code for all items this annotation type can be applied to.
		@hibernate.set table="`AnnotationTypeItems`" lazy="true"
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-element column="`item_type`" type="int"
	*/
	public Set<Integer> getItemTypes()
	{
		if (itemTypes == null)
		{
			itemTypes = new HashSet<Integer>();
		}
		return itemTypes;
	}
	void setItemTypes(Set<Integer> itemTypes)
	{
		this.itemTypes = itemTypes;
	}
	
	public static final int MAX_OPTION_LENGTH = 255;
	private Map<String, String>  options;
	/**
		Additional options depending on the value type.
		@hibernate.map table="`AnnotationTypeOptions`" lazy="true"
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-index column="`name`" type="string" length="255" not-null="true"
		@hibernate.collection-element column="`value`" type="string" length="255" not-null="true"
	*/
	public Map<String, String> getOptions()
	{
		if (options == null)
		{
			options = new HashMap<String, String>();
		}
		return options;
	}
	void setOptions(Map<String, String> options)
	{
		this.options = options;
	}

	private ParameterValueData<?> enumerationValues;
	/**
		The allowed values if this is an enumeration.
		@hibernate.many-to-one column="`enumerationvalues_id`" not-null="false" cascade="all" unique="true"
	*/
	public ParameterValueData<?> getEnumerationValues()
	{
		return enumerationValues;
	}
	public void setEnumerationValues(ParameterValueData<?> enumerationValues)
	{
		this.enumerationValues = enumerationValues;
	}

	private Set<ExperimentData> experiments;
	/**
		This is the inverse end.
		@see ExperimentData#getExperimentalFactors()
		@hibernate.set table="`ExperimentalFactors`" lazy="true" inverse="true"
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-many-to-many column="`experiment_id`" class="net.sf.basedb.core.data.ExperimentData"
	*/
	Set<ExperimentData> getExperiments()
	{
		return experiments;
	}
	void setExperiments(Set<ExperimentData> experiments)
	{
		this.experiments = experiments;
	}
	
	private boolean isProtocolParameter;
	/**
		If this annotation type is intended to be used as a protocol
		parameter or as a regular annotation.
		@since 2.2
		@hibernate.property column="`is_parameter`" type="boolean" not-null="true"
	*/
	public boolean isProtocolParameter()
	{
		return isProtocolParameter;
	}
	public void setProtocolParameter(boolean isProtocolParameter)
	{
		this.isProtocolParameter = isProtocolParameter;
	}
	
	private Set<ProtocolData> protocols;
	/**
		This is the inverse end.
		@see ProtocolData#getParameters()
		@hibernate.set table="`ProtocolParameters`" lazy="true" inverse="true"
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-many-to-many column="`protocol_id`" 
			class="net.sf.basedb.core.data.ProtocolData"
		@since 2.2
	*/
	Set<ProtocolData> getProtocols()
	{
		return protocols;
	}
	void setProtocols(Set<ProtocolData> protocols)
	{
		this.protocols = protocols;
	}
	
	private Set<AnnotationTypeCategoryData> categories;
	/**
		Get the set that manages the categories of this annotation type.
		@hibernate.set table="`CategorisedAnnotationTypes`" lazy="true"
		@hibernate.collection-key column="`annotationtype_id`"
		@hibernate.collection-many-to-many column="`category_id`" class="net.sf.basedb.core.data.AnnotationTypeCategoryData"
	*/
	public Set<AnnotationTypeCategoryData> getCategories() 
	{
		if (categories==null)
		{
			categories= new HashSet<AnnotationTypeCategoryData>();
		}
		return categories;
	}
	
	void setCategories(Set<AnnotationTypeCategoryData> categories) 
	{
		this.categories = categories;
	}

}
