/*
	$Id$

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;

import java.util.List;


/**
	This interface is used by the plugin to access the different parameters
	attached to its job and configuration.
	
	@base.modified $Date$
	@author Enell, Nicklas
	@version 2.0
*/
public interface ParameterValues
{
	/**
		Get the plugin that we are executing/configuring.
		@return A PluginDefinition or null if not known
		@since 2.2
	*/
	public PluginDefinition getPluginDefinition();

	/**
		Get the job that we are executing/configuring.
		@return A Job or null if not known
		@since 2.2
	*/
	public Job getJob();

	/**
		Get the configuration that we are using/configuring.
		@return A PluginConfiguration or null if not known
		@since 2.2
	*/
	public PluginConfiguration getPluginConfiguration();
	
	/**
		Get the ID of the job or configuration the parameters 
		comes from.
	*/
	public int getId();

	/**
		Get a single value for a specific parameter. If the parameter has more then one value
		only the first value is returned. Use {@link #getValues(String)} to return
		all values as a list.
		@param name the name of the parameter
		@return The value or null if no parameter with that name is found
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to read the value
		@throws BaseException If there is another error
	*/
	public Object getValue(String name)
		throws PermissionDeniedException, BaseException;
	
	/**
		Get a <code>List</code> of values for a specific parameter.
		@param name The name of the parameter
		@return The values or null if no parameter with that name is found
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to read the values
		@throws BaseException If there is another error
	*/
	public List<?> getValues(String name)
		throws PermissionDeniedException, BaseException;
	
	/**
		Set a single value for a parameter. The core will store the value at 
		the appropriate place in the database. Ie. as a configuration value to
		either the plugin or job configuration. If the configuration settings are
		write protected because the plugin received an execute command the
		settings will not be changed. If the value is null, but the parameter
		type doesn't allow null values, the default value is used. If the default
		value is null, an exception will be thrown.

		@param name The name of the parameter
		@param type The type of the value
		@param value The value to be set
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the value is invalid according
			to the settings in the type
		@throws BaseException If there is another error
	*/
	public <T> void setValue(String name, ParameterType<T> type, T value)
		throws PermissionDeniedException, InvalidDataException, BaseException;
	
	/**
		Set a list of values for a parameter. The core will store the values at 
		the appropriate place in the database. Ie. as a configuration value to
		either the plugin or job configuration. If the configuration settings are
		write protected because the plugin received an execute command the
		settings will not be changed.

		@param name The name of the parameter
		@param type The type of the value
		@param values A list containing the values
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If any of the values is invalid according
			to the settings in the type
		@throws BaseException If there is another error
	*/
	public <T> void setValues(String name, ParameterType<T> type, List<T> values)
		throws PermissionDeniedException, InvalidDataException, BaseException;
}
