/*
 $Id: BioAssaySetUtil.java 4515 2008-09-11 20:08:14Z jari $

 Copyright (C) 2006 Johan Enell, Nicklas Nordborg, Martin Svensson

 This file is part of BASE - BioArray Software Environment.
 Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.

 BASE is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.basedb.util;

import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.util.jep.ChannelFunction;
import net.sf.basedb.util.jep.ExtraValueFunction;
import net.sf.basedb.util.jep.InReporterListFunction;
import net.sf.basedb.util.jep.Jep;
import net.sf.basedb.util.jep.NotInReporterListFunction;
import net.sf.basedb.util.jep.PositionFunction;
import net.sf.basedb.util.jep.RawFunction;
import net.sf.basedb.util.jep.ReporterFunction;
import net.sf.basedb.util.jep.ScoreFunction;

import java.util.List;
import java.util.Set;
import java.util.TreeSet;

/**
	This is a utility class for {@link net.sf.basedb.core.BioAssaySet}.

	@base.modified $Date: 2008-09-11 22:08:14 +0200 (Thu, 11 Sep 2008) $
	@author Johan
	@version 2.0
*/
public class BioAssaySetUtil
{
	/**
		Find the annotation values for a given annotation type on a bioassay.
		This method will only check annotations, primary and inherited, on the
		raw bioassays that the bioassay is a child of. 
		<p> 
		Since a bioassay may
		have more than one parent raw bioassay and each annotation may have
		multiple values it is possible that the set contains multiple values.
		Each value is only present once regardless of how many times it appears
		in the annotations.
		
		@param dc The DbControl for database access
		@param bioAssay The bioassay to get the annotations for
		@param annotationType The annotation type of the annotations
		@return A set containing all values
	*/
	public static Set<?> getAnnotationValues(DbControl dc, BioAssay bioAssay, AnnotationType annotationType)
	{
		ItemQuery<RawBioAssay> rbaQuery = bioAssay.getRawBioAssays();
		rbaQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		Set<Object> allValues = new TreeSet<Object>();
		if (bioAssay.isAnnotated())
		{
			List<Annotation> all = bioAssay.getAnnotationSet().findAnnotations(dc, annotationType, true);
			for (Annotation a : all)
			{
				if (a != null) allValues.addAll(a.getValues());
			}
		}
		for (RawBioAssay rb : rbaQuery.list(dc))
		{
			if (rb.isAnnotated())
			{
				List<Annotation> all =  rb.getAnnotationSet().findAnnotations(dc, annotationType, true);
				for (Annotation a : all)
				{
					if (a != null) allValues.addAll(a.getValues());
				}
			}
		}
		return allValues;
	}

	/**
		Find the values of parent properties for a given bioassay. This method
		starts by loading all raw bioassays that are parents to the bioassay.
		For each raw bioassay it uses the {@link Metadata#getPropertyValue(DbControl, Object, String)} 
		method to find the value of the given property for that raw bioassay.
		<p> 
		Since a bioassay may have more than one parent raw bioassay and it is possible 
		that the set contains multiple values. Each value is only present once 
		regardless of how many times it appears among the parents.
		<p>
		Examples of property values:
		<ul>
		<li>spots: to get the number of spots in the raw bioassay
		<li>software.name: to get the name of the software used to analyse the image files
		<li>scan.hybridization.creationEvent.eventDate: to get the date when the hybridization
			was created
		</ul>
		
		@param dc The DbControl for database access
		@param bioAssay The bioassay to get the annotations for
		@param property The property to get the value for
		@return A set containing all values
	*/
	public static Set<Object> getParentProperties(DbControl dc, BioAssay bioAssay, String property)
	{
		ItemQuery<RawBioAssay> rbaQuery = bioAssay.getRawBioAssays();
		rbaQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		Set<Object> allValues = new TreeSet<Object>();
		Metadata<RawBioAssay> meta = Metadata.getInstance(RawBioAssay.class);
		for (RawBioAssay rb : rbaQuery.list(dc))
		{
			Object value = meta.getPropertyValue(dc, rb, property);
			if (value == null) value = "Unknown";
			allValues.add(value);
		}
		return allValues;
	}
	
	/**
		Parse a JEP expression and create a restrictions that is suitable to
		be used in a dynamic query. In addition to the standard functions defined
		by JEP (see {@link Jep#getFunctions()}) this method defines the following functions:
		<ul>
		<li>all functions defined by the {@link #createJepExpression(DbControl, BioAssaySet, 
			String, boolean)} method
		<li>inList(list): Checks if the reporter is included in the specified list
		<li>notInList(list): Checks if the reporter isn't included in the specified list
		</ul>
	 	@param dc DbControl to use when accessing the database. 
	 	@param bioAssaySet The current bioassay set. Must not be null.
	 	@param formula The formula to parse. Must not be null.
	 	@param allowRawIfMultipleMappings TRUE if 'raw' function with merged data, FALSE otherwise.
	 	@return A restriction to be used on a query.
	 	@throws InvalidDataException  If any of required parameters are null.
	 	@throws BaseException If creating the restriction fails.
		@see net.sf.basedb.util.jep.Jep
	*/
	public static Restriction createJepRestriction(DbControl dc, BioAssaySet bioAssaySet, 
		String formula, boolean allowRawIfMultipleMappings)
		throws InvalidDataException, BaseException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		if (formula == null) throw new InvalidUseOfNullException("formula");
		RawFunction raw = new RawFunction(bioAssaySet.getMaxRawMappingsForSpot() == 1 || 
			allowRawIfMultipleMappings);
		ChannelFunction ch = new ChannelFunction();
		ReporterFunction reporter = new ReporterFunction();
		ScoreFunction score = new ScoreFunction(dc);
		ExtraValueFunction xtra = new ExtraValueFunction(dc, bioAssaySet);
		InReporterListFunction inList = new InReporterListFunction(dc);
		NotInReporterListFunction notInList = new NotInReporterListFunction(dc);
		return Jep.formulaToRestriction(formula, raw, ch, reporter, score, xtra, inList, notInList);
	}

	/**
		Parse a JEP expression and create a restrictions that is suitable to
		be used in a dynamic query. In addition to the standard functions defined
		by JEP (see {@link Jep#getFunctions()}) this method defines the following functions:
		<ul>
		<li>raw(property): The value of the specified raw data property. The use of this function
			can be denied if the bioassayset contains spots that are mapped to more than one 
			raw data spot. The <code>allowRawIfMultipleMappings</code> parameter controls
			this.
		<li>rep(property): The value of the specified reporter property
		<li>pos: The position of a spot. Note that this is not the same as raw(position)
		<li>ch(n): The intensity in channel n
		<li>score(list): The score of the {@link net.sf.basedb.core.data.ReporterData} in the 
			specified {@link net.sf.basedb.core.ReporterList} (numeric or external ID)
		<li>xtra(extraValue): The value of the specified {@link net.sf.basedb.core.ExtraValue} 
			(numeric ID or the external ID of the {@link net.sf.basedb.core.ExtraValueType})
		</ul>
	  	@param dc The DbControl to use when accessing the database. 
	 	@param bioAssaySet The bioassay set this expression should be used with. Must not be null.
	 	@param formula The formula that should be parsed. Must not be null.
	 	@param allowRawIfMultipleMappings TRUE if 'raw' function with merged data, FALSE otherwise. 
	 	@return An expression to be used with a query.
	 	@throws InvalidDataException If any of required parameters are null.
	 	@throws BaseException If creating the expression fails.
		@see net.sf.basedb.util.jep.Jep
	*/
	public static Expression createJepExpression(DbControl dc, BioAssaySet bioAssaySet, 
		String formula, boolean allowRawIfMultipleMappings)
		throws InvalidDataException, BaseException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		if (formula == null) throw new InvalidUseOfNullException("formula");
		RawFunction raw = new RawFunction(bioAssaySet.getMaxRawMappingsForSpot() == 1 || 
			allowRawIfMultipleMappings);
		ChannelFunction ch = new ChannelFunction();
		PositionFunction pos = new PositionFunction();
		ReporterFunction reporter = new ReporterFunction();
		ScoreFunction score = new ScoreFunction(dc);
		ExtraValueFunction xtra = new ExtraValueFunction(dc, bioAssaySet);
		return Jep.formulaToExpression(formula, raw, ch, pos, reporter, score, xtra);
	}
	
}
