/**
	$Id:ActionFactory.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	An action factory is an object which knows how to create 
	{@link Action}:s. Action factories are part of an extension
	(see {@link Extension#getActionFactory()}).
	
	<p>
	Action factory implementations must be thread-safe, since a 
	single instance of the factory may have to serve multiple requests
	at the same time.
	
	<p>
	Note 1! A single factory instance may be used by more than one 
	extension. This is at the control of the client application.
	
	<p>
	Note 2! The BASE web-client uses the {@link XmlLoader} to 
	define extensions. The <code>XmlLoader</code> always create a
	separate factory instance for each extension. The <code>XmlLoader</code>
	also requires that the factories has a public no-argument constructor.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public interface ActionFactory<A extends Action>
{

	/**
		This method is called once for each request/use of an
		extension and have two purposes:
		
		<ul>
		<li>The factory should decide if the extension should be enabled or
			not. For example, the factory may check the permissions of the
			logged in user and determine that they are inadequate. The boolean
			return value determines if the extension is enabled or disabled.
			
		<li>Initialise the context with resources that the actions may need.
			With the BASE web-client this means that it is possible to
			add scripts or stylesheets that is needed by the extension.
			See {@link net.sf.basedb.clients.web.extensions.JspContext}.
		</ul>
		
		@param context The current invokation context
		@return TRUE if the extension should be enabled, 
			FALSE if the extension should be disabled
	*/
	public boolean prepareContext(InvokationContext<? super A> context);
	
	/**
		This method may be called one or several times for each request. 
		This is decided by the extension point. If, for example, the extension
		point is a pure single-item extension point then this method is probably
		only called once. If the extension point is a per-item extension point in
		a list context, then this method may be called once for every item in the list.
		The context parameter contains all information about the context of
		the extension point, including the current item, if any.
		
		@param context The current invokation context
		@return An array of actions that should be added to the extension point.
			Returns null or an empty array if there are no actions in the current context.
	*/
	public A[] getActions(InvokationContext<? super A> context);
	
}
