/**
	$Id: ASTFunNodeConverter.java 4515 2008-09-11 20:08:14Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep.convert;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.util.jep.JepExpressionFunction;
import net.sf.basedb.util.jep.JepRestrictionFunction;

import org.nfunk.jep.ASTFunNode;
import org.nfunk.jep.Operator;
import org.nfunk.jep.function.PostfixMathCommandI;

/**
	Converts a function node to an expression or restriction. Function
	nodes are used by JEP to represent both operators (+, -, *, ==, !=, etc.)
	and built-in or extended functions (ln, exp, sqrt, etc.).

	@author nicklas
	@version 2.4
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (Thu, 11 Sep 2008) $
*/
public class ASTFunNodeConverter
	extends AbstractJepConversionFunction<ASTFunNode>
{
	
	public ASTFunNodeConverter()
	{}
	
	/*
		From the JepConversionFunction interface
		-------------------------------------------
	*/
	/**
		Get the {@link PostfixMathCommandI} object from the function node
		and delegate conversion to {@link JepExpressionFunction} if that interface
		is implemented. Otherwise check the {@link ConverterFactory} for
		a registered converter and use it.
		@throws BaseException If the node can't be converted
	*/
	@SuppressWarnings("unchecked")
	@Override
	public Expression toExpression(ASTFunNode node)
	{
		Expression e = null;
		PostfixMathCommandI function = node.getPFMC();
		Operator operator = node.getOperator();
		int numAllowed = function.getNumberOfParameters();
		if (numAllowed != -1 && numAllowed != node.jjtGetNumChildren())
		{
			throw new BaseException("Invalid number of expressions for '" + node.getName() + 
				"' function: " + node.jjtGetNumChildren());
		}
		if (function instanceof JepExpressionFunction)
		{
			e = ((JepExpressionFunction)function).toExpression(node);
		}
		else
		{
			JepConversionFunction converter = getConverter(function, operator);
			if (converter != null) 
			{
				e = converter.toExpression(node);
			}
			else
			{
				throw new BaseException("Unsupported JEP function: " + node);
			}
		}
		return e;
	}
	/**
		Get the {@link PostfixMathCommandI} object from the function node
		and delegate conversion to {@link JepRestrictionFunction} if that interface
		is implemented. Otherwise check the {@link ConverterFactory} for
		a registered converter and use it.
		
		Get the {@link Operator} object from the function node
		and check the {@link ConverterFactory} for
		a registered converter and use it.
		@throws BaseException If the node can't be converted
	*/
	@SuppressWarnings("unchecked")
	@Override
	public Restriction toRestriction(ASTFunNode node)
	{
		Restriction r = null;
		PostfixMathCommandI function = node.getPFMC();
		Operator operator = node.getOperator();
		int numAllowed = function.getNumberOfParameters();
		if (numAllowed != -1 && numAllowed != node.jjtGetNumChildren())
		{
			throw new BaseException("Invalid number of expressions for '" + node.getName() + 
				"' function: " + node.jjtGetNumChildren());
		}
		if (function instanceof JepRestrictionFunction)
		{
			r = ((JepRestrictionFunction)function).toRestriction(node);
		}
		else
		{
			JepConversionFunction converter = getConverter(function, operator);
			if (converter != null) 
			{
				r = converter.toRestriction(node);
			}
			else
			{
				throw new BaseException("Unsupported JEP function: " + node);
			}
		}
		return r;
	}
	// -------------------------------------------

	/**
		Get a converter by first checking the if one exists for the given function,
		and then for the given operator symbol.
		@return The converter or null if none is found
	*/
	private JepConversionFunction getConverter(PostfixMathCommandI function, Operator operator)
	{
		JepConversionFunction converter = ConverterFactory.getConverterByClass(function);
		if (converter == null) converter = ConverterFactory.getConverter(operator.getSymbol());
		return converter;
	}
	
}
