/*
	$Id: OneTimeFix.java 4636 2008-11-17 08:35:52Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.install;

import java.io.FileWriter;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import org.hibernate.mapping.Table;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.User;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;

/**
	This class is used for one-time fixes that are needed to be
	executed manually before or after an updgrade.

	@author Nicklas
	@version 2.7.2
	@base.modified $Date: 2008-11-17 09:35:52 +0100 (Mon, 17 Nov 2008) $
*/
public class OneTimeFix
{
	public static void main(String[] args)
	{
		
		if (args.length == 0) 
		{
			showUsage();
			return;
		}
		try
		{
			String cmd = args[0];		
			if ("lowess_warn".equals(cmd))
			{
				String login = getOption(args, "-u", "root");
				String pwd = getOption(args, "-p", "root");
				boolean flag = getOption(args, "-f", null) != null;
				
				String msg = 
					"This script looks for analysed data that has been normalized with\n" +
					"the LOWESS plug-in. This plug-in has had a serious bug that caused\n" +
					"incorrect normalization values to be calculated. The bug has been\n" +
					"fixed in BASE 2.7.2, but all data that has been normalized with\n" +
					"the LOWESS plug-in before this fix must be redone, including any\n" +
					"analysis steps following the normalization. For more information\n" +
					"about the bug see: http://base.thep.lu.se/ticket/1077\n\n" +
					"This script will find and display all experiments and bioassay sets\n" +
					"that has used the LOWESS normalization plug-in. It is not possible to\n" +
					"detect if the normalization was done before or after the fix was\n" +
					"applied. Use the '-f' option when executing this script to tag all\n" +
					"affected experiments and bioassay sets with 'FIX LOWESS' in their names.\n\n" +
					"Press ENTER to continue or CTRL+C to abort";
				System.out.println(msg);
				waitForEnter();
				showDbInfo(null);
				System.out.println("Starting BASE. Please wait...");
				Application.start(false);
				SessionControl sc = Application.newSessionControl(null, null, null);
				sc.login(login, pwd, null, false);
				showLowessBioassaySets(sc, flag);
				sc.logout();
				sc.close();
				Application.stop();
			}
			else if ("utf8".equals(cmd))
			{
				boolean execute = getOption(args, "-x", null) != null;
				boolean force = getOption(args, "-f", null) != null;
				String output = getOption(args, "-o", "convert-to-utf8.sql");

				showDbInfo(null);
				if (!isMySQL())
				{
					System.out.println("This script only works with MySQL.");
					return;
				}
				
				String msg = 
					"This script is used for converting the character set used\n" +
					"by the BASE databases to UTF-8. Only run this script if you\n" +
					"are sure that conversion is needed. The script will first try\n" +
					"to find out if your database needs conversion or not.\n";
				System.out.println(msg);

				String dbCreateSql = getCreateSql();
				boolean isUtf8 = dbCreateSql.contains("utf8");
				msg = "======================================================\n" +
					dbCreateSql + 
					"\n======================================================\n\n";
				if (!isUtf8)
				{
					msg += 
						"YOUR DATABASE NEEDS TO BE CONVERTED SINCE IT USES\n" +
						"A DIFFERENT CHARACTER SET THAN UTF-8.\n";
				}
				else
				{
					msg += 
						"YOUR DATABASE IS ALREADY USING UTF-8 AND DOESN'T HAVE\n" +
						"TO BE CONVERTED.\n";
				}
				if (execute)
				{
					msg += "Press\n   ENTER to convert the database";
					if (isUtf8) msg += " (not recommended)";
				}
				else
				{
					msg += "Press\n   ENTER to save SQL script to '" + 
						output + "'";
				}
				msg += "\nor\n   CTRL-C to abort.";
				
				System.out.println(msg);
				waitForEnter();
				
				if (isUtf8 && execute)
				{
					System.out.println(
						"======================================================\n" +
						"YOUR DATABASE IS ALREADY USING UTF-8.\n" +
						"ARE YOU SURE YOU WANT TO RUN THIS SCRIPT?\n" +
						"======================================================\n" +
						"Press ENTER to continue or CTRL-C to abort.");
					waitForEnter();
				}
				
				System.out.println("Starting UTF-8 conversion/script generation. Please wait...");
				PrintWriter out = new PrintWriter(new FileWriter(output));
				convertToUtf8(execute, force, out);
				out.flush();
				out.close();
			}
			else
			{
				showUsage();
			}
		}
		catch (Throwable t)
		{
			System.out.println("\n");
			t.printStackTrace();
		}
	}

	private static void showUsage()
	{
		System.out.println("Usage: onetimefix.sh cmd <options>");
		System.out.println("generic options: ");
		System.out.println("    -u <login>    : Login name for the root user account; (root)");
		System.out.println("    -p <password> : Password for the root user account; (root)");
		System.out.println("cmd=lowess_warn: ");
		System.out.println("    -f            : Tag names of experiments with 'FIX LOWESS'");
	}

	private static void showDbInfo(ProgressReporter progress)
	{
		final Properties properties = System.getProperties();
		final String jdbcDriver = Config.getString("db.driver");
		String driverVersion = "unknown version";
		try
		{
			final Driver driver = (Driver)Class.forName(jdbcDriver).newInstance();
			driverVersion = "version " + driver.getMajorVersion() + "." + driver.getMinorVersion();
		}
		catch (Throwable t)
		{}
		
		String message = 
			"--System information-----------------------------\n" + 
			"BASE     : " + Application.getMajorVersion() + "." +
				Application.getMinorVersion() + "." + 
				Application.getMaintenanceVersion() + "\n" + 
			"Database : " + Config.getString("db.url") + "\n" + 
			"Dialect  : " + Config.getString("db.dialect") + "\n" + 
			"JDBC     : " + jdbcDriver + "; " + driverVersion + "\n" +
			"Java     : " + properties.getProperty("java.runtime.name") + "; " +
				properties.getProperty("java.runtime.version") + "; " +
				properties.getProperty("java.vendor") + "\n" + 
			"OS       : " + properties.getProperty("os.name") +"; " +
				properties.getProperty("os.arch") + "; " +
				properties.getProperty("os.version") +"\n" +
			"-------------------------------------------------";
		if (progress != null)
		{
			progress.display(0, message);
		}
		else 
		{
			System.out.println(message);
		}
	}
	
	public static String getOption(String[] args, String option, String defaultValue)
	{
		String value = defaultValue;
		for (int i = 0; i < args.length; ++i)
		{
			if (option.equals(args[i]))
			{
				if (i+1 < args.length && !args[i+1].startsWith("-"))
				{
					value = args[i+1];
				}
				else
				{
					value = option;
				}
				return value;
			}
		}
		return value;
	}
	
	private static void showLowessBioassaySets(SessionControl sc, boolean flag)
	{
		DbControl dc = sc.newDbControl();
		int totalBas = 0;
		int totalExp = 0;
		try
		{
			ItemQuery<Experiment> expQuery = Experiment.getQuery();
			expQuery.include(Include.ALL);
			expQuery.order(Orders.asc(Hql.property("name")));
			
			for (Experiment exp : expQuery.list(dc))
			{
				User owner = exp.getOwner();
				ItemQuery<BioAssaySet> basQuery = exp.getBioAssaySets();
				basQuery.order(Orders.asc(Hql.property("name")));
				
				boolean hasPrintedExp = false;
				
				for (BioAssaySet bas : basQuery.list(dc))
				{
					Job job = bas.getTransformation().getJob();
					if (job != null)
					{
						PluginDefinition plugin = job.getPluginDefinition();
						if (plugin != null)
						{
							if (plugin.getClassName().equals("net.sf.basedb.plugins.LowessNormalization"))
							{
								if (!hasPrintedExp)
								{
									String msg = "\nExperiment: " + exp.getName() + 
										"; owner: " + owner.getName() + 
										" \n--------------------------------";
									System.out.println(msg);
									hasPrintedExp = true;
									totalExp++;
									if (flag && !exp.getName().startsWith("FIX LOWESS"))
									{
										String newName = "FIX LOWESS: " + exp.getName();
										exp.setName(StringUtil.trimString(newName, Experiment.MAX_NAME_LENGTH));
									}
								}
								System.out.println("  " + bas.getName() + "; id: " + bas.getId());
								if (flag && !bas.getName().startsWith("FIX LOWESS"))
								{
									String newName = "FIX LOWESS: " + bas.getName();
									bas.setName(StringUtil.trimString(newName, BioAssaySet.MAX_NAME_LENGTH));
								}
								totalBas++;
							}
						}
					}
				}
			}
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		System.out.println("--------------------------");
		if (totalBas == 0)
		{
			System.out.println("Could not find any experiments that has used the LOWESS plug-in.");
		}
		else
		{
			System.out.println("Found " +totalBas + " LOWESS calculations in " + totalExp + " experiments");
		}
	}
	
	/**
		Get a JDBC {@link Connection} to the current BASE database.
		@since 2.9
	*/
	private static Connection getConnection()
		throws SQLException
	{
		String dbUrl = Config.getString("db.url");
		String dbUser = Config.getString("db.username");
		String dbPassword = Config.getString("db.password");
		return DriverManager.getConnection(dbUrl, dbUser, dbPassword);
	}
	
	/**
		Checks if the database is MySQL.
		@since 2.9
	*/
	private static boolean isMySQL()
		throws Exception
	{
		Connection c = null;
		try
		{
			c = getConnection();
			DatabaseMetaData metaData = c.getMetaData();
			String dbName = metaData.getDatabaseProductName();
			return dbName.contains("MySQL");
		}
		finally
		{
			if (c != null) c.close();
		}
	}
	
	/**
		Get the SQL statement that was used to create the 
		current BASE database.
		@since 2.9
	*/
	private static String getCreateSql()
		throws Exception
	{
		Connection c = null;
		Statement s = null;
		ResultSet rs = null;
		String createSql = null;
		try
		{
			c = getConnection();
			DatabaseMetaData metaData = c.getMetaData();
			String dbName = metaData.getDatabaseProductName();
			String sql = "SHOW CREATE DATABASE `" + c.getCatalog() + "`";
			s = c.createStatement();
			rs = s.executeQuery(sql);
			rs.next();
			createSql = "Q: " + sql + "\nA: " + rs.getString("Create Database");
		}
		finally
		{
			if (rs != null) rs.close();
			if (s!= null) s.close();
			if (c != null) c.close();
		}
		return createSql;
	}
	
	/**
		Get the SQL statement that was used to create a given table.
		@since 2.9
	*/
	private static String getCreateTableSql(Statement s, String catalog, String table)
		throws Exception
	{
		ResultSet rs = null;
		String sql = null;
		try
		{
			rs = s.executeQuery("SHOW CREATE TABLE `" + catalog + "`.`" + table + "`");
			rs.next();
			sql = rs.getString("Create Table");
		}
		finally
		{
			if (rs != null) rs.close();
		}
		return sql;
	}
	
	/**
		Convert a BASE database to use UTF-8 for all tables. This only works
		with MySQL. 
		@param execute FALSE to only generate the SQL statements, TRUE to 
			actually execute them on the database server
		@param force TRUE to force conversion of tables that seems to
			not need conversion
		@param out The generated SQL statements are written to this
			writer
		@since 2.9
	*/
	private static void convertToUtf8(boolean execute, boolean force, PrintWriter out)
		throws Exception
	{
		String dynamicDb = Config.getString("db.dynamic.catalog");
		List<String> sqlList = new LinkedList<String>();
		Connection c = null;
		Statement s = null;
		try
		{
			c = getConnection();
			String mainDb = c.getCatalog();
			DatabaseMetaData metaData = c.getMetaData();
			
			generateConvertSql(metaData, mainDb, force, true, sqlList);
			generateConvertSql(metaData, dynamicDb, force, false, sqlList);
			
			if (execute) s = c.createStatement();
			for (String sql : sqlList)
			{
				if (out != null) out.println(sql + ";");
				if (execute) 
				{
					System.out.println(sql);
					long t1 = System.currentTimeMillis();
					int count = s.executeUpdate(sql);
					long t2 = System.currentTimeMillis();
					long t = t2 - t1;
					String time = t < 10000 ? t + " milliseconds" : (int)(t/1000) + " seconds";
					System.out.println(count + " rows affected (" + time + ")\n");
				}
			}
		}
		finally
		{
			if (s!= null) s.close();
			if (c != null) c.close();
		}
	}
	
	/**
		Generates SQL conversion script for all tables in a given
		database (catalog). The is method will generate SQL like this:
		<pre class="code">
ALTER TABLE &lt;table-name&gt;
CONVERT TO CHARACTER SET utf8, 
DEFAULT CHARACTER SET utf8;
</pre>

		<ul>
		<li>If the table is already using UTF-8 it is skipped, 
			unless forceConvert=TRUE
		<li>If the table doesn't have string columns, the following can happen:
			<ol>
			<li>forceConvert=TRUE: The table is converted
			<li>forceConvert=FALSE and changeDefaultForAllTables=TRUE:
				Only the default character set is changed. The <code>CONVERT TO...</code> 
				part of the query is skipped.
			<li>forceConvert=FALSE and changeDefaultForAllTables=FALSE</code>:
				Nothing is done to the table.
			</ol>
		</ul>

		@param metaData The metadata object for the database
		@param catalog The name of the database to convert
		@param forceConvert TRUE to generate SQL for converting tables 
			that seems to not need conversion
		@param changeDefaultForAllTables TRUE to also change the default
			character set for tables that has no string columns
		@param sqlList
		@throws Exception
		@since 2.9
	 */
	private static void generateConvertSql(DatabaseMetaData metaData, String catalog, 
			boolean forceConvert, boolean changeDefaultForAllTables, List<String> sqlList)
		throws Exception
	{
		ResultSet rs = null;
		Statement s = null;
		try
		{
			s = metaData.getConnection().createStatement();
			// Load info about all tables in the given DB.
			rs = metaData.getTables(catalog, null, null, new String[] { "TABLE" });
			while (rs.next())
			{
				Table tbl = new Table();
				tbl.setCatalog(rs.getString("TABLE_CAT"));
				tbl.setName(rs.getString("TABLE_NAME"));
				TableInfo info = new TableInfo(tbl, metaData);
				
				// Check if the table has at least one string-type column
				boolean hasStringColumn = false;
				for (ColumnInfo col : info.getColumns())
				{
					if (col.isStringType())
					{
						hasStringColumn = true;
						break;
					}
				}
				
				// Check if the table already uses UTF-8
				String createSql = getCreateTableSql(s, tbl.getCatalog(), tbl.getName());
				boolean isUtf8 = createSql.contains("utf8");
				
				// Generate conversion script
				String msg = "Table '" + tbl.getCatalog() + "." + tbl.getName() + ": ";
				if (!isUtf8 || forceConvert)
				{
					String sql = "ALTER TABLE `" + tbl.getCatalog() + "`.`" + tbl.getName() + "` " +
						"DEFAULT CHARACTER SET utf8";
					if (hasStringColumn || forceConvert)
					{
						sql += ", CONVERT TO CHARACTER SET utf8";
						msg += "Need conversion to UTF-8.";
					}
					else if (changeDefaultForAllTables)
					{
						msg += "Has no string-type columns. Change default character set to UTF-8.";
					}
					else
					{
						sql = null;
						msg += "Has no string-type columns. No change needed.";
					}
					if (sql != null) sqlList.add(sql);
				}
				else
				{
					msg += "Uses UTF-8 already. No conversion needed.";
				}
				System.out.println(msg);
			}
			rs.close();
		}
		finally
		{
			if (rs != null) rs.close();
		}
		sqlList.add("ALTER DATABASE `" + catalog + "` DEFAULT CHARACTER SET utf8");
	}

	public static void waitForEnter()
	{
		try
		{
			System.in.read();
			System.in.skip(System.in.available());
		}
		catch (Exception ex)
		{}
	}

}

