/**
	$Id $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.File;
import java.io.FileNotFoundException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.jasper.JasperException;
import org.apache.jasper.JspCompilationContext;
import org.apache.jasper.compiler.JDTCompiler;
import org.apache.jasper.servlet.JspServletWrapper;

/**
	Class that enabled compilation of custom JSP pages that are 
	part of an extension that uses classes from the extension JAR
	file (eg. a JAR file in the /WEB-INF/extensions directory). 
	<p>
	
	To make this compiler work the following is required:
	<ul>
	<li>The JSP file must have a '.xjsp' extension.
	<li>This compiler class must be installed in Tomcat's /lib
		directory.
	<li>The '.xjsp' extensions must be mapped to the regular JSP servlet
		with the 'compilerClass' parameter set to this class.
	</ul>
	
	Note! This feature is still experimental. It has only been tested with
	some versions of Tomcat 6.0.x. This class depends on internal implementation 
	details of Tomcat, and may or may not work with future or older versions 
	of Tomcat.
	
	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class XJspCompiler 
	extends JDTCompiler
{
	
	/**
		The location where the extension system is looking for extensions.
	*/
	public static final String EXTENSIONS_URL = "/WEB-INF/extensions";
	
	/**
		The location where the extensions resources are extracted.
	*/
	public static final String RESOURCES_URL = "/extensions";

	/**
		Path to JSP must match: /extensions/[jar-file-name]/[jsp-path].xjsp
		group(1) = JAR name, group(2) = JSP path
	*/
	private static final Pattern PATH_MATCH = 
		Pattern.compile(RESOURCES_URL + "/([^/]+)/(.+\\.xjsp)");
    
	// If this JSP is an extensions JSP
	private boolean isXJsp;
	// The extensions JAR file
	private File extensionsJar;
	// The last modified time of the currently loaded extensions JAR file
	private long lastModified;
	
	public XJspCompiler()
	{}
	
	/**
		From the JDTCompiler class and it's superclasses
		------------------------------------------------
	*/
	/**
		Initialises the compiler. The JSP name is matched against the 
		required pattern and the name of the extensions JAR file is
		extracted from the request path.
	*/
	@Override
	public void init(JspCompilationContext ctxt, JspServletWrapper jsw) 
	{
		String jspName = ctxt.getJspFile();
		Matcher m = PATH_MATCH.matcher(jspName);
		if (m.matches())
		{
			String jarName = m.group(1);
			String jarPath = ctxt.getRealPath(EXTENSIONS_URL + "/" + jarName);
			extensionsJar = new File(jarPath);
			isXJsp = extensionsJar.exists();
        }
		if (log.isDebugEnabled())
		{
			log.debug("Initialising XJSP compiler for file: " + jspName + 
				"; JAR=" + extensionsJar + "; isXJsp=" + isXJsp);
		}
		super.init(ctxt, jsw);
	}

	/**
		Called when a JSP page needs recompiling. We check if the extension JAR file
		has been modified and update the class loader if it has.
	*/
	@Override
	protected void generateClass(String[] smap) 
		throws FileNotFoundException, JasperException, Exception 
	{
		if (log.isDebugEnabled())
		{
			log.debug("Generate class for file: " + ctxt.getJspFile() + 
				"; JAR=" + extensionsJar + "; isXJsp=" + isXJsp);
		}
		setContextClassLoader();
		super.generateClass(smap);
	}

	/**
		Checks if the extensions JAR file has been modified. If not,
		calls super.isOutDated(checkClass)
	*/
	@Override
	public boolean isOutDated(boolean checkClass)
	{
		if (isXJsp && lastModified != extensionsJar.lastModified()) 
		{
			return true;
		}
		return super.isOutDated(checkClass);
	}
	// ---------------------------------------------
	
	/**
		Sets the context class loader if this JSP page is an
		extension JSP page and if the extensions JAR file has been
		modified since the last time the class loader was set.
		This method does nothing if the JSP file is not an extensions
		JSP file, or if the JAR file hasn't changed since the last
		use.
	*/
	protected void setContextClassLoader()
		throws Exception
	{
		if (!isXJsp || lastModified == extensionsJar.lastModified()) return;

		if (log.isDebugEnabled())
		{
			log.debug("Setting class loader for file: " + ctxt.getJspFile() + 
				"; JAR=" + extensionsJar + "; isXJsp=" + isXJsp);
		}

		URLClassLoader classLoader = getExtensionClassLoader();
		ctxt.setClassLoader(classLoader);
		lastModified = extensionsJar.lastModified();
	}
	
	/**
		Get a class loader that can load classes from the current web application
		as well as the extensions JAR file. The class loader for the current
		web application can be found in JspRuntimeContext.getParentClassLoader().
		The extensions class loader is loaded from JarClassLoader in BASE.
	*/
	protected URLClassLoader getExtensionClassLoader()
		throws Exception
	{
		ClassLoader webAppLoader = getWebAppClassLoader();

		// Get a reference to the JarClassLoader.getInstance(String, boolean) method
		Class<?> jarClass = Class.forName("net.sf.basedb.util.JarClassLoader", true, webAppLoader);
		Method getInstance = jarClass.getMethod("getInstance", String.class, boolean.class);

		// Call JarClassLoader.getInstance(jarPath, true)
		ClassLoader jarLoader = (ClassLoader)getInstance.invoke(null, extensionsJar.getAbsolutePath(), true);
		return new URLClassLoader(new URL[0], jarLoader);
	}
	
	/**
		Get the class loader for the web application this JSP is
		located in. Default implementation returns JspRuntimeContext.getParentClassLoader()
	*/
	protected ClassLoader getWebAppClassLoader()
	{
		return ctxt.getRuntimeContext().getParentClassLoader();
	}

	
}
