/*
	$Id: IntensityCalculatorPlugin.java 4569 2008-10-08 07:18:21Z nicklas $

	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.IntensityCalculator;
import net.sf.basedb.util.IntensityCalculatorUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
   This plugin provides methods to calculate the intensity for a root bioassayset 
   using different formulas. It needs the following parameters:

   <pre>
   String              : name; The name of the new bioassayset
   Experiment          : experiment; The experiment we are working on
   List&gt;RawBioAssay&lt;   : rawBioAssays; The raw bioassays to create bioassays for
     (must be part of the experiment)
   Formula             : formula; The formula to use to for calculating intesities
   </pre>

   @author Gregory, Nicklas
   @version 2.0
   @base.modified $Date: 2008-10-08 09:18:21 +0200 (Wed, 08 Oct 2008) $
*/
public class IntensityCalculatorPlugin
	extends AbstractPlugin
	implements InteractivePlugin, SignalTarget
{

	/**
	   Plugin description
	*/
	private static final About about = 
	        new AboutImpl
		(
			"Formula intensity calculator",
			"This plugin is used to calculate intensities from raw data creating an initial root bioassay set. "+
			"The intensities are calculated using Formula objects registered for the experiment's raw data type. " +
			"Thus, the functionality of this plugin can be extended by adding more formula objects to the database.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);

	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.BIOASSAYSET, GuiContext.Type.LIST));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private static final StringParameterType nameType = 
		new StringParameterType(BioAssaySet.MAX_NAME_LENGTH, "New bioassayset", true);
	private static final PluginParameter<String> nameParameter = new PluginParameter<String>(
		"name", "Bioassay set name", "The name of the root bioassayset", nameType);

	private ItemParameterType<Experiment> experimentType;
	private PluginParameter<Experiment> experimentParameter;

	private ItemParameterType<RawBioAssay> rawBioAssaysType;
	private PluginParameter<RawBioAssay> rawBioAssaysParameter;

	private ItemParameterType<Formula> formulaType;
	private PluginParameter<Formula> formulaParameter;

	private RequestInformation configureJob;
	private ThreadSignalHandler signalHandler;
	
	/**
	   Constructor should be empty. 
	*/
	public IntensityCalculatorPlugin()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	/**
	   Returns the main plugin type this plugin belongs to.
	   @return {@link net.sf.basedb.core.plugin.Plugin.MainType#INTENSITY}.
	*/
	public Plugin.MainType getMainType()
	{
		return Plugin.MainType.INTENSITY;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request use access to Experiment:s and read access to RawBioAssay:s, ArrayDesign:s, 
	 	Formulas:s, Job:s and and Reporter:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			// Grant write access to RawBioAssays and read access to ArrayDesign, File and Reporter
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.RAWBIOASSAY, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FORMULA, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	@SuppressWarnings("unchecked")
	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) 
		{
			signalHandler.setWorkerThread(null);
		}
		DbControl dc = sc.newDbControl();
		try
		{
			// name parameter
			String name = (String)job.getValue("name");
			if (name == null) name = "New root bioassay set";
	
			// experiment parameter
			Experiment experiment = (Experiment)job.getValue("experiment");
			if (experiment == null)
			{
				throw new InvalidUseOfNullException("experiment");
			}
			
			// To avoid disconnection/reconnection issues
			experiment = Experiment.getById(dc, experiment.getId());
	
			// rawBioAssays parameter
			List<RawBioAssay> sources = (List<RawBioAssay>)job.getValues("rawBioAssays");
			if (sources == null || sources.size() == 0)
			{
				throw new InvalidUseOfNullException("sources");
			}
			for (RawBioAssay rba : sources)
			{
				dc.reattachItem(rba);
			}

			// Formula parameter
			Formula formula = (Formula)job.getValue("formula");
			formula = Formula.getById(dc, formula.getId());
	
			// intensity calculation
			RawDataType rdt = experiment.getRawDataType();
			
			// Check that the formula is an intensity expression
			if (formula.getFormulaType() != Formula.Type.INTENSITY_EXPRESSION)
			{
				String msg = "The formula '" + formula.getName() + "' can't be used for intensity " +
					"calculation, since it is a '" + formula.getFormulaType() + "', not a '" + 
					Formula.Type.INTENSITY_EXPRESSION + "'";
				throw new InvalidDataException(msg);
			}
			
			// Check that the formula can be used with this raw data type
			if (rdt != formula.getRawDataType())
			{
				String msg = "The formula '" + formula.getName() + "' can " +
					"only be used with the " + rdt + " raw data type, not with the " +
					formula.getRawDataType() + " raw data type.";
				throw new InvalidDataException(msg);
			}

			int jobId = job.getId();
			Job thisJob = Job.getById(dc, jobId);

			checkInterrupted();
			BioAssaySet rootBas = null;
			if (IntensityCalculatorUtil.checkSameArrayDesign(sources, false))
			{
				/*
					All raw bioassays use the same array design, use Expressions
					This method should in theory be faster since it leaves most of the
					work on the database. For MySQL it doesn't seem to matter. It takes 
					more or less the same amount of time as the alternate method, but
					on Postgres we get a huge performance increase (about 5-10 times faster).
				*/
				Expression[] expressions = IntensityCalculatorUtil.createJepExpressions(dc, rdt, formula.getFormulas().toArray(new String[formula.getNumFormulas()]));
				rootBas = IntensityCalculatorUtil.createRootBioAssaySet(dc, experiment, sources, thisJob, expressions, progress);
			}
			else
			{
				// Different array designs, use IntensityCalculator
				IntensityCalculator iCalc = 
					IntensityCalculatorUtil.createJepIntensityCalculator(dc, rdt, 
							formula.getFormulas().toArray(new String[formula.getNumFormulas()]));
				rootBas = IntensityCalculatorUtil.createRootBioAssaySet(dc, experiment, sources, thisJob, iCalc, progress);
			}
			rootBas.setName(name);
			Transformation t = rootBas.getTransformation();
			t.setName(formula.getName());
			dc.commit();
			if (progress != null) progress.display(100, "Done");
			response.setDone("Done");
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Check that the current experiment stores it's raw data in
		the database.
	*/
	public String isInContext(GuiContext context, Object item)
	{
		DbControl dc = sc.newDbControl();
		try
		{
			Experiment e = Experiment.getById(dc, sc.getCurrentContext(Item.EXPERIMENT).getId());
			if (!e.getRawDataType().isStoredInDb())
			{
				return "Raw data for raw data type '" + e.getRawDataType() + "' is not stored in the database.";
			}
			else
			{
				e.checkPermission(Permission.USE);
				ItemQuery<RawBioAssay> query = e.getRawBioAssays();
				query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
				query.restrict(Restrictions.gt(Hql.property("spots"), Expressions.integer(0)));
				if (query.count(dc) == 0)
				{
					return "The experiment doesn't have any raw bioassay with raw data imported into the database";
				}
			}
		}
		catch (Throwable ex)
		{}
		finally
		{
			if (dc != null) dc.close();
		}
		return null;
	}
	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will not ask for
		any parameters.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		the parameters mentioned in the class documentation.
	*/
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters();
		}
		return requestInformation;
	}
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN}, 
		and {@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	@SuppressWarnings("unchecked")
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				List<Throwable> errors = validateRequestParameters(configureJob.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				Experiment experiment = (Experiment)request.getParameterValue("experiment");
				DbControl dc = sc.newDbControl();
				int numSpots = 0;
				try
				{
					experiment = Experiment.getById(dc, experiment.getId());
					if (!experiment.getRawDataType().isStoredInDb())
					{
						throw new BaseException("Raw data for raw data type '" + experiment.getRawDataType() +
							"' is not stored in the database.");
					}
					List<RawBioAssay> rawBioAssays = (List<RawBioAssay>)request.getParameterValues("rawBioAssays");
					for (RawBioAssay rba : rawBioAssays)
					{
						if (!experiment.isUsing(rba))
						{
							throw new InvalidDataException("Raw bioassay '" + rba.getName() + 
								"' is not part of the experiment '" + experiment.getName()+"'");
						}
						if (rba.getNumDbSpots() == 0)
						{
							throw new BaseException("Raw bioassay '" + rba.getName() + 
								"' has not imported raw data into the database");
						}
						numSpots += rba.getNumDbSpots();
					}
				}
				finally
				{
					if (dc != null) dc.close();
				}
				
				Formula formula = (Formula)request.getParameterValue("formula");
				if (formula.getFormulaType() != Formula.Type.INTENSITY_EXPRESSION)
				{
					throw new InvalidDataException("The selected formula is not an intensity expression formula: " + 
						formula.getName());
				}
				if (formula.getRawDataType() != experiment.getRawDataType())
				{
					throw new InvalidDataException("The selected formula '"+formula.getName()+
						"' uses a different raw data type than the experiment '" + experiment.getName()+"'");
				}
				
				storeValue(job, request, nameParameter);
				storeValue(job, request, experimentParameter);
				storeValues(job, request, rawBioAssaysParameter);
				storeValue(job, request, formulaParameter);
				
				Job.ExecutionTime execTime = Job.ExecutionTime.SHORT;
				// TODO - this should be configurable in some way
				int shortLimit = 1000000;
				int mediumLimit = shortLimit * 10;
				if (numSpots > mediumLimit)
				{
					execTime = Job.ExecutionTime.LONG;
				}
				else if (numSpots > shortLimit)
				{
					execTime = Job.ExecutionTime.MEDIUM;
				}
				
				response.setDone("Job configuration complete", execTime);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
			ex.printStackTrace();
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	
	private RequestInformation getConfigureJobParameters()
	{
		if (configureJob == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			experimentType = new ItemParameterType<Experiment>(Experiment.class, null, true, 1, null);
			experimentParameter = new PluginParameter<Experiment>(
				"experiment", "Experiment", "The experiment we are working with", experimentType);
			
			parameters.add(experimentParameter);
			parameters.add(nameParameter);
			
			DbControl dc = sc.newDbControl();
			try
			{
				Experiment experiment = (Experiment)job.getValue("experiment");
				if (experiment == null)
				{
					int experimentId = sc.getCurrentContext(Item.EXPERIMENT).getId();
					experiment = Experiment.getById(dc, experimentId);
				}
				
				ItemQuery<RawBioAssay> rawBioAssayQuery = experiment.getRawBioAssays();
				rawBioAssayQuery.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
				rawBioAssayQuery.restrict(Restrictions.gt(Hql.property("spots"), Expressions.integer(0)));
				rawBioAssayQuery.order(Orders.asc(Hql.property("name")));
				List<RawBioAssay> rawBioAssays = new ArrayList<RawBioAssay>(rawBioAssayQuery.list(dc));
				rawBioAssaysType = new ItemParameterType<RawBioAssay>(RawBioAssay.class, null, true, 0, rawBioAssays);
				rawBioAssaysParameter = new PluginParameter<RawBioAssay>(
					"rawBioAssays", "Raw bioassays", 
					"Select the raw bioassays to use when creating the root bioassay set.\n\n" +
					"NOTE!\n Only raw bioassays with raw data imported into the database can " +
					"be selected.",
					rawBioAssaysType);
				
				parameters.add(rawBioAssaysParameter);
				
				ItemQuery<Formula> formulaQuery = Formula.getQuery(Formula.Type.INTENSITY_EXPRESSION, 
						experiment.getRawDataType());
				formulaQuery.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
				formulaQuery.order(Orders.asc(Hql.property("name")));
				List<Formula> formulas = new ArrayList<Formula>(formulaQuery.list(dc));
				formulaType = new ItemParameterType<Formula>(Formula.class, null, true, 1, formulas);
				formulaParameter = new PluginParameter<Formula>(
					"formula", "Formula", "Select the expression to use for calculating the intensities",
					formulaType);
				parameters.add(formulaParameter);
			}
			finally
			{
				if (dc != null) dc.close();
			}			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Intensity calculation options",
				"Select which raw bioassays to use and a formula for the calculations",
				parameters
			);
		}
		return configureJob;
	}

}
