/*
	$Id: RawDataFlatFileImporter.java 4654 2008-11-20 12:13:33Z nicklas $

	Copyright (C) 2005, 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.FileType;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataBatcher;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.UnsupportedFileFormatException;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AnnotationSetterPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.ConstantMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.FlatFileParser.Line;
import net.sf.basedb.core.Job;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
	A plugin that imports rawdata from a flatfile.
	
	@base.modified $Date: 2008-11-20 13:13:33 +0100 (Thu, 20 Nov 2008) $
	@author Enell
	@version 2.0
*/
public class RawDataFlatFileImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample, AnnotationSetterPlugin
{
	
	private static final About about = 
		new AboutImpl
		(
			"Raw data importer",
			"This plugin is used to import raw data from a simple text file.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);
	
	private static final Set<GuiContext> guiContexts = 
		Collections.unmodifiableSet(new HashSet<GuiContext>(
				Arrays.asList(
					GuiContext.item(Item.RAWBIOASSAY),
					GuiContext.item(Item.EXPERIMENT)
				)
			));	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	private static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	private static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);
	
	private static final PluginParameter<String> positionColumnMapping = new PluginParameter<String>(
		"positionColumnMapping",
		"Position",
		"Mapping that picks the spot's position from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the POSITION method " +
		"for identifying features. If the raw data is not connected to an array design " +
		"position values will automatically be generated if no mapping has been specified. In all" +
		"cases, position value must be unique." +
		"Mapping example: \\Position\\",
		optionalColumnMapping
		);

	
	private static final PluginParameter<String> blockColumnMapping = new PluginParameter<String>(
		"blockColumnMapping",
		"Block",
		"Mapping that picks the spot's block number from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the COORDINATES method " +
		"for identifying features. In all other cases, the value is stored as is." +
		"For example: \\Block\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> columnColumnMapping = new PluginParameter<String>(
		"columnColumnMapping",
		"Column",
		"Mapping that picks the spot's column coordinate from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the COORDINATES method " +
		"for identifying features. In all other cases, the value is stored as is." +
		"For example: \\Column\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> rowColumnMapping = new PluginParameter<String>(
		"rowColumnMapping",
		"Row",
		"Mapping that picks the spot's row coordinate from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the COORDINATES method " +
		"for identifying features. In all other cases, the value is stored as is." +
		"For example: \\Row\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> metaGridXColumnMapping = new PluginParameter<String>(
		"metaGridXColumnMapping",
		"MetaGridX",
		"Mapping that picks the spot's meta-Grid X-axis from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the COORDINATES method " +
		"for identifying features. In all other cases, the value is stored as is." +
		"For example: \\Meta grid X\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> metaGridYColumnMapping = new PluginParameter<String>(
		"metaGridYColumnMapping",
		"MetaGridY",
		"Mapping that picks the spot's meta-Grid Y-axis from the data columns. This column is only " +
		"used when the raw data is connected to an array design which uses the COORDINATES method " +
		"for identifying features. In all other cases, the value is stored as is." +
		"For example: \\Meta grid Y\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> xColumnMapping = new PluginParameter<String>(
		"xColumnMapping",
		"X",
		"Mapping that picks the spot's physical X coordinate from the data columns. " +
		"For example: \\X\\",
		optionalColumnMapping
		);
	
	private static final PluginParameter<String> yColumnMapping = new PluginParameter<String>(
		"yColumnMapping",
		"Y",
		"Mapping that picks the spot's physical Y coordinate from the data columns. " +
		"For example: \\Y\\",
		optionalColumnMapping
		);

	private static final PluginParameter<String> reporterIdColumnMapping = new PluginParameter<String>(
		"reporterIdColumnMapping",
		"Reporter ID",
		"Mapping that picks the 'External ID' of the spot's reporter from the data columns. " +
		"For example: \\ID\\",
		requiredColumnMapping
		);

	private static final PluginParameter<String> featureIdColumnMapping = new PluginParameter<String>(
			"featureIdColumnMapping",
			"Feature ID",
			"Mapping that picks the spot's feature ID from the data columns. " +
			"This column is only used when the raw data is connected to an array design " +
			"which uses the FEATURE_ID method for identifying features. The value is not saved " +
			"to the database." +
			"For example: \\Feature ID\\",
			optionalColumnMapping
		);
	
	
	protected static final PluginParameter<String> missingReporterErrorParameter = new PluginParameter<String>(
		"missingReporterError",
		"Reporter not found",
		"How to handle errors that are caused by a reporter not beeing present in the datbase. If not specified the " +
			"default error handling is used.\n\n"+
			"null = Import the data but set the reporter to null\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
					Arrays.asList( new String[] { "null", "skip", "fail"} ))
		);
	
	protected static final PluginParameter<String> featureMismatchErrorParameter = new PluginParameter<String>(
		"featureMismatchError",
		"Feature mismatch",
		"How to handle errors that are caused by the data not matching the feature on the array design. " +
			"This can for example be caused by invalid block and/or spot coordinates or a different reporter. " +
			"If not specified the default error handling is used.\n\n"+
			"smart = SKIP lines that have no feature on the array design, " +
			"ACCEPT lines that has a feature with a NULL reporter, " +
			"FAIL lines that has a feature with different reporter\n" +
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, "smart", false, 1, 0, 0, 
				Arrays.asList( new String[] { "smart", "skip", "fail"} ))
		);

	protected static final PluginParameter<String> duplicateFeatureErrorParameter = new PluginParameter<String>(
		"duplicateFeatureError",
		"Duplicate feature",
		"How to handle errors that are caused by more than data spot matching the same " +
			"feature on the array design. This can for example be caused by invalid block and/or " +
			"spot coordinates or using an incorrect feature or reporter ID column. " +
			"If not specified the default error handling is used.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);
	
	
	public static final String COMMAND_CONFIGURE_PARSER = "config_parser";
	
	private List<PluginParameter<String>> allColumnMappings;
	
	private RequestInformation configureParser;
	private RequestInformation configureJob;
	private RequestInformation configureRawDataType;
	
	private StringParameterType rawDataTypeType;
	private PluginParameter<String> rawDataTypeParameter;
	
	private DbControl dc;
	private FlatFileParser ffp;
	private NumberFormat numberFormat;
	private boolean nullIfException;
	private boolean continueAfterError;
	private boolean transactionPerFile;
	private FileIterator fileIterator;
	private long totalFileSize;
	
	private RawBioAssay rawBioAssay;
	private RawDataBatcher batcher;
	private FileSetMember rawDataMember;

	private Mapper featureIdMapper;
	private Mapper reporterMapper;
	private Mapper positionMapper;
	private Mapper blockMapper;
	private Mapper columnMapper;
	private Mapper rowMapper;
	private Mapper metaGridXMapper;
	private Mapper metaGridYMapper;
	private Mapper xMapper;
	private Mapper yMapper;
	private Map<RawDataProperty, Mapper> extendedMappers;

	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return true;
	}
	public boolean requiresConfiguration()
	{
		return true;
	}
	/**
	 	Request write access to RawBioAssay:s and read access to ArrayDesign:s, File:s, 
	 	 Reporter:s, Platform:s and DataFileType:s
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.RAWBIOASSAY, null, EnumSet.of(Permission.READ, Permission.WRITE)));
			permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.WRITE, Permission.CREATE)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLATFORM, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FILETYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLUGINCONFIGURATION, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Returns null if the item is a {@link RawBioAssay} of the correct
		{@link RawDataType} and doesn't already have spots.
		@throws PermissionDeniedException If the raw bioasssay already has raw data
			or if the logged in user doesn't have write permission
	*/
	public String isInContext(GuiContext context, Object item)
	{
		if (item == null)
		{
			return "The object is null";
		}
		else if (!context.isCorrectBasicItem(item))
		{
			return "The oject is not a " + context.getItem() + ": " + item;
		}
		else
		{
			Experiment exp = null;
			RawBioAssay rba = null;
			RawDataType rdt = null;
			if (item instanceof RawBioAssay)
			{
				rba = (RawBioAssay)item;
				rdt = rba.getRawDataType();
				if (rba.getNumDbSpots() > 0)
				{
					throw new PermissionDeniedException("The raw bioassay already has data.");
				}
				rba.checkPermission(Permission.WRITE);
			}
			else
			{
				exp = (Experiment)item;
				rdt = exp.getRawDataType();
				ItemQuery<RawBioAssay> query = getRawBioAssayQuery(exp);
				DbControl dc = sc.newDbControl();
				try
				{
					if (query.count(dc) == 0)
					{
						return "The experiment doesn't have any raw bioassays with files waiting for import";
					}
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			String rawDataType = (String)configuration.getValue("rawDataType");
			if (!rdt.getId().equals(rawDataType))
			{
				return "Unsupported raw data type: " + rdt.getName();
			}
			else if (!rdt.isStoredInDb())
			{
				return "Raw data for raw data type '" + rdt + "' is not stored in the database";
			}
		}
		return null;
	}

	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will ask for
		what type of rawdata to import.
		The {@link #COMMAND_CONFIGURE_PARSER} command will ask for
		parser regular expressions and column mappings.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a file and the {@link RawBioAssay} where the data should be added.
		
		@param command 
	*/
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigureRawDataTypeParameters();
		}
		else if(command.equals(COMMAND_CONFIGURE_PARSER))
		{
			RawDataType rdt = RawDataTypes.getSafeRawDataType((String)configuration.getValue("rawDataType"));
			requestInformation = getConfigureParserParameters(rdt);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.setPropertyFilter(FileType.getPropertyFilter(FileType.RAW_DATA));
			fileContext.setPropertyFilter(getPrimaryLocationFilter());
		}
		return requestInformation;
	}

	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN}, 
		{@link #COMMAND_CONFIGURE_PARSER} and {@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				List<Throwable> errors = 
					validateRequestParameters(getConfigureRawDataTypeParameters().getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				storeValue(configuration, request, rawDataTypeParameter);
				response.setContinue(COMMAND_CONFIGURE_PARSER);
			}
			else if (command.equals(COMMAND_CONFIGURE_PARSER))
			{
				RawDataType rdt = RawDataTypes.getSafeRawDataType((String)configuration.getValue("rawDataType"));
				RequestInformation ri = getConfigureParserParameters(rdt);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				// Parser settings
				storeValue(configuration, request, headerRegexpParameter);
				storeValue(configuration, request, dataHeaderRegexpParameter);
				storeValue(configuration, request, dataSplitterRegexpParameter);
				storeValue(configuration, request, trimQuotesParameter);
				storeValue(configuration, request, ignoreRegexpParameter);
				storeValue(configuration, request, dataFooterRegexpParameter);
				storeValue(configuration, request, minDataColumnsParameter);
				storeValue(configuration, request, maxDataColumnsParameter);
				storeValue(configuration, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(configuration, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				
				// Column mappings
				storeValue(configuration, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : allColumnMappings)
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(configuration, pp, mapExpression);
				}
				
				response.setDone("Plugin configuration complete");
			}
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Single raw bioassay mode
				storeValue(job, request, ri.getParameter("rawBioAssay"));
				storeValue(job, request, ri.getParameter("file"));
				storeValue(job, request, ri.getParameter("featureIdentification"));
				
				// Experiment mode
				storeValue(job, request, ri.getParameter("experiment"));
				storeValue(job, request, ri.getParameter("dataFileType"));
				storeValue(job, request, ri.getParameter("transactionPerFile"));
				storeValue(job, request, ri.getParameter("continueAfterError"));
				
				// Common
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(job, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				
				// Annotations
				storeValue(job, request, ri.getParameter(COPY_ANNOTATIONS));
				storeValue(job, request, ri.getParameter(OVERWRITE_ANNOTATIONS));
				
				// Error handling parameters
				storeValue(job, request, ri.getParameter(Parameters.LOGFILE_PARAMETER));
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, missingReporterErrorParameter);
				storeValue(job, request, featureMismatchErrorParameter);
				storeValue(job, request, duplicateFeatureErrorParameter);
				storeValue(job, request, stringTooLongErrorParameter);
				storeValue(job, request, invalidUseOfNullErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, numberOutOfRangeErrorParameter);
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				// TODO - maybe check file size to make a better estimate
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}

	// -------------------------------------------
	
	/**
		Return a query that finds annotation types for raw bioassays.
		@since 2.5
	*/
	public ItemQuery<AnnotationType> getAnnotationTypes()
	{
		return AnnotationType.getQuery(Item.RAWBIOASSAY);
	}
	public String getAnnotationMessage()
	{
		return "The selected annotations will be copied to the raw bioassay when raw data is imported to it.";
	}
	
	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	/**
		Prepare for the import. If used from an experiment context
		prepare the list of files/raw bioassays to import.
	*/
	@Override 
	protected void start()
	{
		transactionPerFile = Boolean.TRUE.equals(job.getValue("transactionPerFile"));
		continueAfterError = transactionPerFile && 
			Boolean.TRUE.equals(job.getValue("continueAfterError"));
		Experiment exp = (Experiment)job.getValue("experiment");
		List<RawBioAssay> rawBioAssays = null;
		List<File> files = null;
		if (exp == null)
		{
			File f = (File)job.getValue("file");
			RawBioAssay r = (RawBioAssay)job.getValue("rawBioAssay");
			rawBioAssays = Collections.singletonList(r);
			files = Collections.singletonList(f);
			totalFileSize = f.getSize();
		}
		else
		{
			ItemQuery<RawBioAssay> query = getRawBioAssayQuery(exp);
			DataFileType fileType = (DataFileType)job.getValue("dataFileType");
			rawBioAssays = new ArrayList<RawBioAssay>();
			files = new ArrayList<File>();
			DbControl dc = sc.newDbControl();
			try
			{
				for (RawBioAssay rba : query.list(dc))
				{
					File dataFile = null;
					if (fileType != null)
					{
						FileSet fs = rba.getFileSet();
						FileSetMember member = fs.getMember(fileType);
						if (member != null) dataFile = member.getFile();
					}
					else
					{
						List<File> rawFiles = FileStoreUtil.getGenericDataFiles(dc, rba, FileType.RAW_DATA);
						if (rawFiles != null && rawFiles.size() > 0) dataFile = rawFiles.get(0);
					}
					if (dataFile != null)
					{
						rawBioAssays.add(rba);
						files.add(dataFile);
						totalFileSize += dataFile.getSize();
					}
				}
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		this.fileIterator = new FileIterator(rawBioAssays, files);
	}
	
	/**
		We will continue if:
		<ul>
		<li>The file is not of the correct format, eg. t = {@link UnsupportedFileFormatException}.
		<li>The 'Continue after error' parameter is true
		</ul>
	*/
	protected boolean continueWithNextFileAfterError(Throwable t)
	{
		return continueAfterError || t instanceof UnsupportedFileFormatException;
	}
	
	@Override
	protected Iterator<File> getFileIterator()
	{
		return fileIterator;
	}
	
	@Override
	protected long getTotalFileSize()
	{
		return totalFileSize;
	}
	
	/**
		Prepare for importing the next file. Eg. 
		create a {@link DbControl} (if needed) and a {@link RawDataBatcher}. 
		Load the {@link RawBioAssay} that has been specified.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
		throws BaseException
	{
		super.begin(ffp);
		if (transactionPerFile || dc == null || dc.isClosed()) 
		{
			dc = sc.newDbControl();
		}

		rawBioAssay = RawBioAssay.getById(dc, fileIterator.getCurrentRawBioAssay().getId());
		rawDataMember = FileStoreUtil.setGenericDataFile(dc, rawBioAssay, 
			FileType.RAW_DATA, DataFileType.GENERIC_RAW_DATA, fileIterator.getCurrentFile());
		
		PluginConfiguration cfg = getCurrentConfiguration(dc);
		if (cfg != null && cfg.isAnnotated() && !Boolean.FALSE.equals(job.getValue("copyAnnotations")))
		{
			boolean overwrite = Boolean.TRUE.equals(job.getValue("overwriteAnnotations"));
			rawBioAssay.getAnnotationSet().copyFrom(cfg, overwrite);
		}
		
		FeatureIdentificationMethod fiMethod = null;
		if (job.getValue("featureIdentification") != null)
		{
			try
			{
				fiMethod = FeatureIdentificationMethod.valueOf((String)job.getValue("featureIdentification"));
			}
			catch (Exception ex)
			{}
		}
		
		batcher = rawBioAssay.getRawDataBatcher(fiMethod, job.getJob());
		this.ffp = ffp;
		this.numberFormat = ffp.getDefaultNumberFormat();
		this.nullIfException = "null".equals(job.getValue("numberFormatError"));
		
		// Set up error handling
		if (rawBioAssay.hasArrayDesign())
		{
			String method = (String)job.getValue("featureMismatchError");
			if (method != null)
			{
				addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
				boolean isSmart = "smart".equals(method);
				batcher.skipInsertIfMissingFeature(isSmart);
				batcher.acceptInsertIfNullReporterOnFeature(isSmart);
			}
		}
		else
		{
			String method = (String)job.getValue("missingReporterError");
			batcher.useNullIfReporterNotFound("null".equals(method));
			if (method != null)
			{
				addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
			}
		}
		// Error handling for "duplicate feature"
		String method = (String)job.getValue("duplicateFeatureError");
		if (method != null)
		{
			addErrorHandler(ItemAlreadyExistsException.class, new SimpleErrorHandler("skip".equals(method)));
		}

	}

	/**
		Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		boolean cropStrings = ("crop".equals(job.getValue("stringTooLongError")));
		// Mapper that always return null
		Mapper nullMapper = new ConstantMapper((String)null);
		reporterMapper = getMapper(ffp, (String)configuration.getValue("reporterIdColumnMapping"), 
			cropStrings ? ReporterData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);
		
		featureIdMapper = getMapper(ffp, (String)configuration.getValue("featureIdColumnMapping"), 
			cropStrings ? FeatureData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);

		positionMapper = getMapper(ffp, (String)configuration.getValue("positionColumnMapping"), null, nullMapper);
		blockMapper = getMapper(ffp, (String)configuration.getValue("blockColumnMapping"), null, nullMapper);
		columnMapper = getMapper(ffp, (String)configuration.getValue("columnColumnMapping"), null, nullMapper);
		rowMapper = getMapper(ffp, (String)configuration.getValue("rowColumnMapping"), null, nullMapper);
		metaGridXMapper = getMapper(ffp, (String)configuration.getValue("metaGridXColumnMapping"), null, nullMapper);
		metaGridYMapper = getMapper(ffp, (String)configuration.getValue("metaGridYColumnMapping"), null, nullMapper);
		xMapper = getMapper(ffp, (String)configuration.getValue("xColumnMapping"), null, nullMapper);
		yMapper = getMapper(ffp, (String)configuration.getValue("yColumnMapping"), null, nullMapper);

		extendedMappers = new HashMap<RawDataProperty, Mapper>();
		for (RawDataProperty ep : rawBioAssay.getRawDataType().getProperties())
		{
			String name = "propertyMapping." + ep.getName();
			Mapper m = getMapper(ffp, (String)configuration.getValue(name), 
				cropStrings && ep.getLength() > 0 ? ep.getLength() : null, null);
			if (m != null) extendedMappers.put(ep, m);
		}
	}
	
	/**
	 Close and commit/rollback the RawDataBatcher and DbControl.
	*/
	@Override
	protected void end(boolean success)
		throws BaseException
	{
		if (rawDataMember != null) rawDataMember.setValid(success, null);

		try
		{
			if (success)
			{
				batcher.close();
				if (transactionPerFile) 
				{
					dc.commit();
					dc = null;
				}
			}
		}
		catch (BaseException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null && transactionPerFile) 
			{
				dc.close();
				dc = null;
			}
			super.end(success);
		}
	}
	
	/**
		Commit/rollback transaction if needed. If everything
		went ok, t=null.
	*/
	@Override
	public String finish(Throwable t)
	{
		try
		{
			if (t == null && dc != null && !dc.isClosed()) dc.commit();
			super.finish(t);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return null;
	}
	
	@Override
	protected void handleData(Data data)
		throws BaseException
	{
		RawData raw = batcher.newRawData();

		String externalId = reporterMapper.getValue(data);
		String featureId = featureIdMapper.getValue(data);
		Integer position = positionMapper.getInt(data);
		if (position != null) raw.setPosition(position);
		raw.setBlock(blockMapper.getInt(data));
		raw.setColumn(columnMapper.getInt(data));
		raw.setRow(rowMapper.getInt(data));
		raw.setMetaGridX(metaGridXMapper.getInt(data));
		raw.setMetaGridY(metaGridYMapper.getInt(data));
		raw.setX(xMapper.getFloat(data));
		raw.setY(yMapper.getFloat(data));

		for (Map.Entry<RawDataProperty, Mapper> entry : extendedMappers.entrySet())
		{
			RawDataProperty ep = entry.getKey();
			Mapper m = entry.getValue();
			raw.setExtended(ep.getName(), ep.parseString(m.getValue(data), numberFormat, nullIfException));
		}
		batcher.insert(raw, externalId, featureId);
	}
	
	@Override
	protected void handleHeader(Line line) 
		throws BaseException
	{
		super.handleHeader(line);
		if (line.name() != null && line.value() != null)
		{
			rawBioAssay.setHeader(line.name(), line.value());
		}
	}
	/**
		Return <code>x spots inserted</code>.
	*/
	protected String getSuccessMessage(int skippedLines)
	{
		StringBuilder msg = new StringBuilder();
		msg.append(batcher.getTotalInsertCount()).append(" spot(s) inserted");
		int numReportersNotFound = batcher.getNumReportersNotFound();
		int numSkippedMissingFeature = batcher.getNumSkippedMissingFeature();
		int numAcceptedWithNullReporterOnFeature = batcher.getNumAcceptedWithNullReporterOnFeature();
		if (numReportersNotFound > 0)
		{
			msg.append("; ").append(numReportersNotFound).append(" spot(s) could not find a reporter");
		}
		if (numAcceptedWithNullReporterOnFeature > 0)
		{
			msg.append("; ").append(numAcceptedWithNullReporterOnFeature);
			msg.append(" spot(s) accepted with null reporter on feature");
		}
		if (numSkippedMissingFeature > 0)
		{
			msg.append("; ").append(numSkippedMissingFeature);
			msg.append(" spot(s) skipped due to missing features");
		}
		if (skippedLines > 0) 
		{
			msg.append("; ").append(skippedLines).append(" spot(s) skipped due to errors");
		}
		return msg.toString();
	}
	// -------------------------------------------

	/**
		Loads the column mappings if they hasn't been loaded already.
		@param rdt the <code>RawDataType</code> that defines the extra 
			properties.
	*/
	private List<PluginParameter<String>> getAllColumnMappings(RawDataType rdt)
	{
		if (allColumnMappings == null)
		{
			allColumnMappings = new ArrayList<PluginParameter<String>>();
			allColumnMappings.add(reporterIdColumnMapping);
			allColumnMappings.add(featureIdColumnMapping);
			allColumnMappings.add(positionColumnMapping);
			allColumnMappings.add(blockColumnMapping);
			allColumnMappings.add(columnColumnMapping);
			allColumnMappings.add(rowColumnMapping);
			allColumnMappings.add(metaGridXColumnMapping);
			allColumnMappings.add(metaGridYColumnMapping);
			allColumnMappings.add(xColumnMapping);
			allColumnMappings.add(yColumnMapping);
			for(RawDataProperty rdp : rdt.getProperties())
			{
				allColumnMappings.add(
					new PluginParameter<String>(
						"propertyMapping."+rdp.getName(),
						rdp.getTitle(),
						rdp.getDescription(),
						optionalColumnMapping
					)
				);
			}
		}
		return allColumnMappings;
	}
	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			// Parameters for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			RawBioAssay currentRaw = (RawBioAssay)job.getValue("rawBioAssay");
			Experiment currentExp = (Experiment)job.getValue("experiment");
			boolean configIsAnnotated = false;
			boolean useMissingReporterParameter = true;
			boolean useFeatureParameters = true;
			FeatureIdentificationMethod defaultFiMethod = FeatureIdentificationMethod.NONE;
			DbControl dc = sc.newDbControl();
			try
			{
				configIsAnnotated = getCurrentConfiguration(dc).isAnnotated();
				if (GuiContext.item(Item.RAWBIOASSAY).equals(context))
				{
					if (currentRaw == null)
					{
						int id = sc.getCurrentContext(Item.RAWBIOASSAY).getId();
						if (id != 0) currentRaw = RawBioAssay.getById(dc, id);
					}
					else
					{
						currentRaw = RawBioAssay.getById(dc, currentRaw.getId());
					}
					if (currentRaw != null)
					{
						useFeatureParameters = currentRaw.hasArrayDesign();
						ArrayDesign design = currentRaw.getArrayDesign();
						if (design != null) defaultFiMethod = design.getFeatureIdentificationMethod();
					}
					useMissingReporterParameter = !useFeatureParameters;
					addRawBioAssayParameters(dc, currentRaw, parameters);
				}
				else
				{
					if (currentExp == null)
					{
						int id = sc.getCurrentContext(Item.EXPERIMENT).getId();
						if (id != 0) currentExp = Experiment.getById(dc, id);
					}
					addExperimentParameters(dc, currentExp, parameters);
				}
			}
			finally 
			{
				if (dc != null) dc.close();
			}
				
			if (useFeatureParameters)
			{
				// Parameter for asking which feature identification method to use
				// Only used if the connected array design has features
				List<String> methods = new ArrayList<String>();
				String description = "Choose which method to use for identifying features: \n";
				if (configuration.getValue("rowColumnMapping") != null && configuration.getValue("columnColumnMapping") != null)
				{
					methods.add("COORDINATES");
					description += "COORDINATES: Use block, meta-grid, row and column coordinates\n";
				}
				if (configuration.getValue("positionColumnMapping") != null)
				{
					methods.add("POSITION");
					description += "POSITION: Use a position number only\n";
				}
				if (configuration.getValue("featureIdColumnMapping") != null)
				{
					methods.add("FEATURE_ID");
					description += "FEATURE_ID: Use the feature ID value (string)\n";
				}
				
				if (methods.size() > 0)
				{
					PluginParameter<String> featureIdentificationParameter = 
						new PluginParameter<String>(
							"featureIdentification",
							"Identify features by",
							description,
							new StringParameterType(255, 
								methods.contains(defaultFiMethod.name()) ? defaultFiMethod.name() : null, 
								false, 1, 0, 0, methods)
						);
					parameters.add(featureIdentificationParameter);
				}
			}
			
			parameters.add(Parameters.charsetParameter(null, null, 
					currentExp != null ? null : (String)configuration.getValue(Parameters.CHARSET_PARAMETER)));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, 
					(String)configuration.getValue(Parameters.DECIMAL_SEPARATOR_PARAMETER)));
			
			// Annotations section
			if (configIsAnnotated)
			{
				parameters.add(annotationSection);
				parameters.add(getCopyAnnotationsParmeter(null, 
					"If selected, annotations are copied from the configuration to the raw " +
					"bioassay we are importing data to.", true));
				parameters.add(getOverwriteAnnotationsParameters(null, null, null));
			}
			
			// Error handling parameters
			parameters.add(errorSection);
			parameters.add(Parameters.logFileParameter(null, null, null));
			parameters.add(defaultErrorParameter);
			
			if (useMissingReporterParameter)
			{
				parameters.add(missingReporterErrorParameter);
			}
			if (useFeatureParameters)
			{
				parameters.add(featureMismatchErrorParameter);
			}
			parameters.add(duplicateFeatureErrorParameter);
			parameters.add(stringTooLongErrorParameter);
			parameters.add(invalidUseOfNullErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(numberOutOfRangeErrorParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select the file that contains the raw data",
				"Here you select which file to import the raw data from.",
				parameters
			);
		}
		return configureJob;
	}	
	
	private void addRawBioAssayParameters(DbControl dc, RawBioAssay currentRaw, List<PluginParameter<?>> parameters)
	{
		// Parameter for getting the current raw bioassay
		PluginParameter<RawBioAssay> rbaParameter = new PluginParameter<RawBioAssay>
		(
			"rawBioAssay",
			"Raw bioassay",
			"The raw bioassay to import raw data to. It must be of the correct raw data " +
			"type and shouldn't have any spots already",
			new ItemParameterType<RawBioAssay>(RawBioAssay.class, currentRaw, true, 1, null)
		);
		parameters.add(rbaParameter);
	
		List<File> rawDataFiles = null;
		if (currentRaw != null)
		{
			try
			{
				rawDataFiles = FileStoreUtil.getGenericDataFiles(dc, currentRaw, FileType.RAW_DATA);
			}
			catch (Throwable t)
			{}
		}
		
		// The raw data file to import from - if a file already hase
		// been attached to the raw bioassay use it as a default choice
		PluginParameter<File> fileParameter = new PluginParameter<File>(
			"file",
			"Raw data file",
			"The file that contains the raw data that you want to import",
			new FileParameterType(rawDataFiles == null || rawDataFiles.isEmpty() ? 
				null : rawDataFiles.get(0), true, 1)
		);
		parameters.add(fileParameter);

	}
	
	private void addExperimentParameters(DbControl dc, Experiment currentExp, List<PluginParameter<?>> parameters)
	{
		// Parameter for getting the current raw bioassay
		PluginParameter<Experiment> expParameter = new PluginParameter<Experiment>
		(
			"experiment",
			"Experiment",
			"The experiment to import raw data to",
			new ItemParameterType<Experiment>(Experiment.class, currentExp, true, 1, null)
		);
		
		Platform platform = null;
		PlatformVariant variant = null;
		if (currentExp != null)
		{
			RawDataType type = currentExp.getRawDataType();
			platform = type.getPlatform(dc);
			variant = type.getVariant(dc);
		}
		FileType rawData = FileType.getById(dc, SystemItems.getId(FileType.RAW_DATA));
		ItemQuery<DataFileType> query = DataFileType.getQuery(platform, variant, 
			Item.RAWBIOASSAY, rawData);
		query.include(Include.ALL);
		query.order(Orders.asc(Hql.property("name")));
		List<DataFileType> fileTypes = query.list(dc);

		PluginParameter<DataFileType> fileTypeParameter = new PluginParameter<DataFileType>
		(
			"dataFileType",
			"File type",
			"The file type to use when looking for files to import. If not selected " +
			"a generic lookup for any raw data file type is used. ",
			new ItemParameterType<DataFileType>(DataFileType.class, null, false, 1, fileTypes)
		);

		PluginParameter<Boolean> transactionParameter = new PluginParameter<Boolean>
		(
			"transactionPerFile",
			"One transaction per file",
			"If this option is enabled, a new transaction is used for each " +
			"raw bioassay to import (recommended). This means that if an import fails, " +
			"you will not loose what has already been done.",
			new BooleanParameterType(true, false)
		);
		
		PluginParameter<Boolean> continueParameter = new PluginParameter<Boolean>
		(
			"continueAfterError",
			"Continue with next file after an error",
			"If this option is enabled the import will continue with the next " +
			"raw bioassay after an error (recommended). This feature requires that " +
			"the 'One transaction per file' option is enabled",
			new BooleanParameterType(true, false)
		);
		
		parameters.add(expParameter);
		parameters.add(fileTypeParameter);
		parameters.add(transactionParameter);
		parameters.add(continueParameter);
	}
	
	private RequestInformation getConfigureRawDataTypeParameters()
	{
		if (configureRawDataType == null)
		{
			// Load raw data types and initialise the rawDataTypeParameter
			Enumeration<String, String> types = new Enumeration<String, String>();
			for (RawDataType rdt : RawDataTypes.getRawDataTypes())
			{
				if (rdt.isStoredInDb())
				{
					types.add(rdt.getId(), rdt.getName());
				}
			}
			rawDataTypeType = new StringParameterType(255, null, true, 1, 0, 0, types);
			rawDataTypeParameter = new PluginParameter<String>
			(
				"rawDataType",
				"Raw data type",
				"The type of raw data that this importer will import.",
				rawDataTypeType
			);

			// Parameters object for CONFIGURE_PLUGIN
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(rawDataTypeParameter);
			
			configureRawDataType = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Select raw data type",
				"This is required since different raw data type have different data columns",
				parameters
			);
		}
		return configureRawDataType;
	}
	
	private RequestInformation getConfigureParserParameters(RawDataType rdt)
	{
		if (configureParser == null)
		{
			// RequestInformation object for CONFIGURE_PARSER
			ArrayList<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(headerRegexpParameter);
			parameters.add(dataHeaderRegexpParameter);
			parameters.add(dataSplitterRegexpParameter);
			parameters.add(trimQuotesParameter);
			parameters.add(ignoreRegexpParameter);
			parameters.add(dataFooterRegexpParameter);
			parameters.add(minDataColumnsParameter);
			parameters.add(maxDataColumnsParameter);
			parameters.add(Parameters.charsetParameter(null, null, null));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, null));

			// Column mappings
			parameters.add(mappingSection);
			parameters.add(complexMappings);
			parameters.addAll(getAllColumnMappings(rdt));

			configureParser = new RequestInformation
			(
				COMMAND_CONFIGURE_PARSER,
				"Parser settings",
				"Enter the regular expressions used to parse the text file " +
				"and the column mappings used to find matching properties for the raw data.",
				parameters
			);

		}
		return configureParser;
	}
	
	/**
		Get a query that returns all raw bioassays which it may be possible to
		import raw data to for the given experiment. The raw bioassays 
		should have no raw data already and they must have at least one
		file attached.
		@param exp The experiment
	*/
	private ItemQuery<RawBioAssay> getRawBioAssayQuery(Experiment exp)
	{
		ItemQuery<RawBioAssay> query = exp.getRawBioAssays();
		query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
		// Only return raw bioassays with no data...
		query.restrict(Restrictions.eq(Hql.property("spots"), Expressions.integer(0)));
		// ... that has files attached
		query.restrict(Restrictions.neq(Hql.property("fileSet"), null));
		return query;
	}

	
	class FileIterator
		implements Iterator<File>
	{
		private List<RawBioAssay> rawBioAssays;
		private List<File> files;
		private int currentIndex;
		
		FileIterator(List<RawBioAssay> rawBioAssays, List<File> files)
		{
			this.rawBioAssays = rawBioAssays;
			this.files = files;
			this.currentIndex = -1;
		}

		/*
			From the Iterator interface
			---------------------------
		*/
		@Override
		public boolean hasNext()
		{
			return rawBioAssays.size() - 1 > currentIndex;
		}

		@Override
		public File next()
		{
			currentIndex++;
			log("Importing data to raw bioassay: " + getCurrentRawBioAssay().getName());
			return getCurrentFile();
		}

		@Override
		public void remove()
		{
			throw new UnsupportedOperationException("remove");
		}
		// -------------------------------------
		
		File getCurrentFile()
		{
			return files.get(currentIndex);
		}
		
		RawBioAssay getCurrentRawBioAssay()
		{
			return rawBioAssays.get(currentIndex);
		}
		
	}
	
}
